from abc import ABC
import warnings
import numpy as np
import mxnet as mx
from .inference_parameters import InferenceParameters
from ..common.config import get_default_device
from ..common.exceptions import InferenceError
from ..util.inference import discover_shape_constants, init_outcomes
from ..models.factor_graph import FactorGraph


class Inference(ABC):
    """
    Abstract class defining an inference method that can be applied to a model.
    An inference method consists of a few components: the applied inference algorithm, the model definition (optionally a definition of posterior
    approximation), the inference parameters.

    :param inference_algorithm: The applied inference algorithm
    :type inference_algorithm: InferenceAlgorithm
    :param graphs: a list of graph definitions required by the inference method. It includes the model definition and necessary posterior approximation.
    :type graphs: [FactorGraph]
    :param observed: A list of observed variables
    :type observed: [Variable]
    :param constants: Specify a list of model variables as constants
    :type constants: {Variable: mxnet.ndarray}
    :param hybridize: Whether to hybridize the MXNet Gluon block of the inference method.
    :type hybridize: boolean
    :param dtype: data type for internal numberical representation
    :type dtype: {numpy.float64, numpy.float32, 'float64', 'float32'}
    :param context: The MXNet context
    :type context: {mxnet.cpu or mxnet.gpu}
    """

    def __init__(self, inference_algorithm, constants=None,
                 hybridize=False, dtype=None, context=None):

        self.dtype = dtype if dtype is not None else np.float32
        self.mxnet_context = context if context is not None else \
            get_default_device()
        self._hybridize = hybridize
        self._graphs = inference_algorithm.graphs
        self._infr_alg = inference_algorithm
        self.params = InferenceParameters(constants=constants,
                                          dtype=self.dtype,
                                          context=self.mxnet_context)
        self._initialized = False

    @property
    def observed_variables(self):
        return self._infr_alg.observed_variables

    @property
    def observed_variable_UUIDs(self):
        return self._infr_alg.observed_variable_UUIDs

    @property
    def observed_variable_names(self):
        return self._infr_alg.observed_variable_names

    @property
    def graphs(self):
        """
        Returns the list of graphs the inference algorithm operates over.
        """
        return self._graphs

    @property
    def inference_algorithm(self):
        """
        Return the reference to the used inference algorithm.
        """
        return self._infr_alg

    def create_executor(self):
        """
        Return a MXNet Gluon block responsible for the execution of the inference method.
        """
        infr = self._infr_alg.create_executor(data_def=self.observed_variable_UUIDs,
                                              params=self.params,
                                              var_ties=self.params.var_ties)
        if self._hybridize:
            infr.hybridize()
        infr.initialize(ctx=self.mxnet_context)
        return infr

    def _initialize_params(self):
        self.params.initialize_params(self._graphs, self.observed_variable_UUIDs)

    def initialize(self, **kw):
        """
        Initialize the inference method with the shapes of observed variables. The inputs of the keyword arguments are the names of the
        variables in the model defintion. The values of the keyword arguments are the data of the corresponding variables (mxnet.ndarray)
        or their shape (tuples).
        """
        if not self._initialized:
            data = [kw[v] for v in self.observed_variable_names]
            if len(data) > 0:
                if not all(isinstance(d, type(d)) for d in data):
                    raise InferenceError("All items in the keywords must be of the same type. Either all shapes or all data objects.")

                if isinstance(data[0], (tuple, list)):
                    data_shapes = {i: d for i, d in zip(self.observed_variable_UUIDs, data)}
                elif isinstance(data[0], mx.nd.ndarray.NDArray):
                    data_shapes = {i: d.shape for i, d in zip(self.observed_variable_UUIDs,
                                                              data)}
                else:
                    raise InferenceError("Keywords not of type mx.nd.NDArray or tuple/list for shapes passed into initialization.")
            shape_constants = discover_shape_constants(data_shapes,
                                                       self._graphs)
            self.params.update_constants(shape_constants)
            self._initialize_params()
            self._initialized = True
        else:
            warnings.warn("Trying to initialize the inference twice, skipping.")
            # TODO: how to handle when initialization called twice

    def run(self, **kw):
        """
        Run the inference method.

        :param \**kwargs: The keyword arguments specify the data for inferenceself. The key of each argument is the name of the corresponding
            variable in model definition and the value of the argument is the data in numpy array format.
        :returns: the samples of target variables (if not spcified, the samples of all the latent variables)
        :rtype: {UUID: samples}
        """
        data = [kw[v] for v in self.observed_variable_names]
        self.initialize(**kw)
        infr = self.create_executor()
        return infr(mx.nd.zeros(1, ctx=self.mxnet_context), *data)

    def set_intializer(self):
        """
        Configure the inference method on how to initialize variables and parameters.
        """
        pass

    def load(self,
             primary_model_file=None,
             secondary_graph_files=None,
             inference_configuration_file=None,
             parameters_file=None,
             mxnet_constants_file=None,
             variable_constants_file=None):
        """
        Loads back everything needed to rerun an inference algorithm. The major pieces of this are the InferenceParameters, FactorGraphs, and
        InferenceConfiguration.

        :param primary_model_file: The file containing the primary model to load back for this inference algorithm.
        :type primary_model_file: str of filename
        :param secondary_graph_files: The files containing any secondary graphs (e.g. a posterior) to load back for this inference algorithm.
        :type secondary_graph_files: [str of filename]
        :param inference_configuration_file: The file containing any inference specific configuration needed to reload this inference algorithm.
            e.g. observation patterns used to train it.
        :type inference_configuration_file: str of filename
        :param parameters_file: These are the parameters of the previous inference algorithm.  These are in a {uuid: mx.nd.array} mapping.
        :type mxnet_constants_file: file saved down with mx.nd.save(), so a {uuid: mx.nd.array} mapping saved in a binary format.
        :param mxnet_constants_file: These are the constants in mxnet format from the previous inference algorithm. These are in a {uuid: mx.nd.array} mapping.
        :type mxnet_constants_file: file saved down with mx.nd.save(), so a {uuid: mx.nd.array} mapping saved in a binary format.
        :param variable_constants_file: These are the constants in primitive format from the previous inference algorithm.
        :type variable_constants_file: json dict of {uuid: constant_primitive}
        """
        primary_model = FactorGraph('graph_0').load_graph(primary_model_file)
        secondary_graphs = [
            FactorGraph('graph_{}'.format(str(i + 1))).load_graph(file)
            for i, file in enumerate(secondary_graph_files)]

        # { current_model_uuid : loaded_uuid}
        self._uuid_map = FactorGraph.reconcile_graphs(
            current_graphs=self.graphs,
            primary_previous_graph=primary_model,
            secondary_previous_graphs=secondary_graphs)
        new_parameters = InferenceParameters.load_parameters(
            uuid_map=self._uuid_map,
            parameters_file=parameters_file,
            variable_constants_file=variable_constants_file,
            mxnet_constants_file=mxnet_constants_file,
            current_params=self.params._params)
        self.params = new_parameters
        self.load_configuration(inference_configuration_file, self._uuid_map)

    def load_configuration(self, config_file, uuid_map):
        """
        Loads relevant inference configuration back from a file. Currently only loads the observed variables UUIDs back in, using the uuid_map
        parameter to store the correct current observed variables.

        :param config_file: The file to save the configuration down into.
        :type config_file: str
        :param uuid_map: A map of previous/loaded model component uuids to their current variable in the loaded graph.
        :type uuid_map: { current_model_uuid : loaded_previous_uuid}
        """
        import json
        with open(config_file) as f:
            configuration = json.load(f)
        # loaded_uuid = [uuid_map[uuid] for uuid in configuration['observed']]

    def save_configuration(self, config_file):
        """
        Saves relevant inference configuration down into a file. Currently only saves the observed variables UUIDs as {'observed': [observed_uuids]}.

        :param config_file: The file to save the configuration down into.
        :type config_file: str
        """
        import json
        with open(config_file, 'w') as f:
            json.dump({'observed': self.observed_variable_UUIDs}, f, ensure_ascii=False)

    def save(self, prefix=None):
        """
        Saves down everything needed to reload an inference algorithm. The two primary pieces of this are the InferenceParameters and FactorGraphs.

        :param prefix: The directory and any appending tag for the files to save this Inference as.
        :type prefix: str , ex. "../saved_inferences/experiment_1"
        """
        prefix = prefix if prefix is not None else "inference"
        self.params.save(prefix=prefix)
        self.save_configuration(prefix + '_configuration.json')
        for i, g in enumerate(self._graphs):
            filename = prefix + "_graph_{}.json".format(i)
            g.save(filename)


class TransferInference(Inference):
    """
    The abstract Inference method for transfering the outcome of one inference
    method to another.

    :param inference_algorithm: The applied inference algorithm
    :type inference_algorithm: InferenceAlgorithm
    :param graphs: a list of graph definitions required by the inference method. It includes the model definition and necessary posterior approximation.
    :type graphs: [FactorGraph]
    :param observed: A list of observed variables
    :type observed: [Variable]
    :param constants: Specify a list of model variables as constants
    :type constants: {Variable: mxnet.ndarray}
    :param hybridize: Whether to hybridize the MXNet Gluon block of the inference method.
    :type hybridize: boolean
    :param dtype: data type for internal numberical representation
    :type dtype: {numpy.float64, numpy.float32, 'float64', 'float32'}
    :param context: The MXNet context
    :type context: {mxnet.cpu or mxnet.gpu}
    """
    def __init__(self, inference_algorithm, infr_params, var_tie=None,
                 constants=None, hybridize=False, dtype=None, context=None):
        self._var_tie = var_tie if var_tie is not None else {}
        self._inherited_params = infr_params
        super(TransferInference, self).__init__(
            inference_algorithm=inference_algorithm, constants=constants,
            hybridize=hybridize, dtype=dtype, context=context)

    def generate_executor(self, **kw):

        data_shapes = [kw[v] for v in self.observed_variable_names]
        if not self._initialized:
            self._initialize_run(self._var_tie, self._inherited_params,
                                 data_shapes)
            self._initialized = True

        infr = self._infr_alg.create_executor(
            data_def=self.observed_variable_UUIDs, params=self.params,
            var_ties=self.params.var_ties)
        if self._hybridize:
            infr.hybridize()
        infr.initialize()
        return infr

    def _initialize_params(self):
        self.params.initialize_with_carryover_params(
            self._graphs, self.observed_variable_UUIDs, self._var_tie,
            init_outcomes(self._inherited_params))



import mxnet as mx
from .grad_loop import GradLoop


class BatchInferenceLoop(GradLoop):
    """
    The class for the main loop for batch gradient-based optimization.
    """

    def run(self, infr_executor, data, param_dict, ctx, optimizer='adam',
            learning_rate=1e-3, max_iter=2000, verbose=False):
        """
        :param infr_executor: The MXNet function that computes the training objective.
        :type infr_executor: MXNet Gluon Block
        :param data: a list of observed variables
        :type data: [mxnet.ndarray]
        :param param_dict: The MXNet ParameterDict for Gradient-based optimization
        :type param_dict: mxnet.gluon.ParameterDict
        :param ctx: MXNet context
        :type ctx: mxnet.cpu or mxnet.gpu
        :param optimizer: the choice of optimizer (default: 'adam')
        :type optimizer: str
        :param learning_rate: the learning rate of the gradient optimizer (default: 0.001)
        :type learning_rate: float
        :param max_iter: the maximum number of iterations of gradient optimization
        :type max_iter: int
        :param verbose: whether to print per-iteration messages.
        :type verbose: boolean
        """
        trainer = mx.gluon.Trainer(param_dict,
                                   optimizer=optimizer,
                                   optimizer_params={'learning_rate':
                                                     learning_rate})
        for i in range(max_iter):
            with mx.autograd.record():
                loss = infr_executor(mx.nd.zeros(1, ctx=ctx), *data)
                loss.backward()
            if verbose:
                print('\rIteration {} logL: {}'.format(i + 1, -loss.asscalar()))
            trainer.step(batch_size=1, ignore_stale_grad=True)



"""This module contains functionality for using functions in MXFusion models.

Submodules
==========

.. autosummary::
    :toctree: _autosummary

    function_evaluation
    gluon_func_eval
    mxfusion_gluon_function
"""

from .mxfusion_gluon_function import MXFusionGluonFunction
from .function_evaluation import FunctionEvaluation



from ..factor import Factor


class Module(Factor):
    """
    Modules are a combined Model/Posterior + Inference algorithm object in MXFusion. They act as Factors and are defined as such during model
    definition, producing Random Variables like a typical Distribution. During inference, instead of the Module having a closed form solution,
    the outside inference algorithm will call the module's internal inference algorithm to perform modular inference.
    """

    def __init__(self, model, inferences):
        """
        Constructor.

        :param model: The model for this module.
        :param inferences: Inferences for this module.
        """
        self._model = model
        self._inferences = inferences

    def create_variable(self):
        """
        Create a variable drawn from this module for use during model building.
        """
        raise NotImplementedError



from .kernel import NativeKernel
from ....variables import Variable
from ....variables import PositiveTransformation
from .....util.customop import broadcast_to_w_samples


class Bias(NativeKernel):
    """
    Bias kernel, which produces a constant value for every entries of the covariance matrix.

    .. math::
       k(x,y) = \\sigma^2

    :param input_dim: the number of dimensions of the kernel. (The total number of active dimensions).
    :type input_dim: int
    :param variance: the initial value for the variance parameter.
    :type variance: float or MXNet NDArray
    :param name: the name of the kernel. The name is used to access kernel parameters.
    :type name: str
    :param active_dims: The dimensions of the inputs that are taken for the covariance matrix computation. (default: None, taking all the dimensions).
    :type active_dims: [int] or None
    :param dtype: the data type for float point numbers.
    :type dtype: numpy.float32 or numpy.float64
    :param ctx: the mxnet context (default: None/current context).
    :type ctx: None or mxnet.cpu or mxnet.gpu
    """
    def __init__(self, input_dim, variance=1., name='bias', active_dims=None,
                 dtype=None, ctx=None):
        super(Bias, self).__init__(input_dim=input_dim, name=name,
                                   active_dims=active_dims, dtype=dtype,
                                   ctx=ctx)
        if not isinstance(variance, Variable):
            variance = Variable(shape=(1,),
                                transformation=PositiveTransformation(),
                                initial_value=variance)
        self.variance = variance

    def _compute_K(self, F, X, variance, X2=None):
        """
        The internal interface for the actual covariance matrix computation.

        :param F: MXNet computation type <mx.sym, mx.nd>.
        :param X: the first set of inputs to the kernel.
        :type X: MXNet NDArray or MXNet Symbol
        :param X2: (optional) the second set of arguments to the kernel. If X2 is None, this computes a square covariance matrix of X. In other words,
            X2 is internally treated as X.
        :type X2: MXNet NDArray or MXNet Symbol
        :param variance: the variance parameter.
        :type variance: MXNet NDArray or MXNet Symbol
        :return: The covariance matrix.
        :rtype: MXNet NDArray or MXNet Symbol
        """
        if X2 is None:
            X2 = X
        return broadcast_to_w_samples(F, variance, X.shape[:-1] +
                                      (X2.shape[-2],))

    def _compute_Kdiag(self, F, X, variance):
        """
        The internal interface for the actual computation for the diagonal.

        :param F: MXNet computation type <mx.sym, mx.nd>.
        :param X: the first set of inputs to the kernel.
        :type X: MXNet NDArray or MXNet Symbol
        :param variance: the variance parameter.
        :type variance: MXNet NDArray or MXNet Symbol
        :return: The covariance matrix.
        :rtype: MXNet NDArray or MXNet Symbol
        """
        return broadcast_to_w_samples(F, variance, X.shape[:-1])


class White(NativeKernel):
    """
    White kernel, which produces a constant value for the diagonal of the covariance matrix.

    .. math::
       K = \sigma^2 I

    :param input_dim: the number of dimensions of the kernel. (The total number of active dimensions).
    :type input_dim: int
    :param variance: the initial value for the variance parameter.
    :type variance: float or MXNet NDArray
    :param name: the name of the kernel. The name is used to access kernel parameters.
    :type name: str
    :param active_dims: The dimensions of the inputs that are taken for the covariance matrix computation. (default: None, taking all the dimensions).
    :type active_dims: [int] or None
    :param dtype: the data type for float point numbers.
    :type dtype: numpy.float32 or numpy.float64
    :param ctx: the mxnet context (default: None/current context).
    :type ctx: None or mxnet.cpu or mxnet.gpu
    """
    def __init__(self, input_dim, variance=1., name='white', active_dims=None,
                 dtype=None, ctx=None):
        super(White, self).__init__(input_dim=input_dim, name=name,
                                    active_dims=active_dims, dtype=dtype,
                                    ctx=ctx)
        if not isinstance(variance, Variable):
            variance = Variable(shape=(1,),
                                transformation=PositiveTransformation(),
                                initial_value=variance)
        self.variance = variance

    def _compute_K(self, F, X, variance, X2=None):
        """
        The internal interface for the actual covariance matrix computation.

        :param F: MXNet computation type <mx.sym, mx.nd>
        :param X: the first set of inputs to the kernel.
        :type X: MXNet NDArray or MXNet Symbol
        :param X2: (optional) the second set of arguments to the kernel. If X2 is None, this computes a square covariance matrix of X. In other words,
            X2 is internally treated as X.
        :type X2: MXNet NDArray or MXNet Symbol
        :param variance: the variance parameter.
        :type variance: MXNet NDArray or MXNet Symbol
        :return: The covariance matrix.
        :rtype: MXNet NDArray or MXNet Symbol
        """
        if X2 is None:
            Imat = F.eye(N=X.shape[-2:-1][0],
                         ctx=self.ctx,
                         dtype=self.dtype)
            Imat = broadcast_to_w_samples(F, Imat, X.shape[:-1] +
                                          X.shape[-2:-1], False)
            return Imat * broadcast_to_w_samples(F, variance, X.shape[:-1] +
                                                 X.shape[-2:-1])
        else:
            return F.zeros(shape=X.shape[:-1] + X2.shape[-2:-1], ctx=self.ctx,
                           dtype=self.dtype)

    def _compute_Kdiag(self, F, X, variance):
        """
        The internal interface for the actual computation for the diagonal of the covariance matrix.

        :param F: MXNet computation type <mx.sym, mx.nd>.
        :param X: the first set of inputs to the kernel.
        :type X: MXNet NDArray or MXNet Symbol
        :param variance: the variance parameter.
        :type variance: MXNet NDArray or MXNet Symbol
        :return: The covariance matrix.
        :rtype: MXNet NDArray or MXNet Symbol
        """
        return broadcast_to_w_samples(F, variance, X.shape[:-1])



import mxnet as mx
from mxnet.operator import CustomOp, CustomOpProp
from .util import parse_string_to_tuple


class BroadcastToWithSamplesOp(CustomOp):
    def __init__(self, isSamples, shape, **kwargs):
        self.isSamples = isSamples
        self.shape = shape
        super(BroadcastToWithSamplesOp, self).__init__(**kwargs)

    def forward(self, is_train, req, in_data, out_data, aux):
        a = in_data[0]
        n_dim = len(self.shape)
        if self.isSamples:
            t_shape = (a.shape[0],) + (1,) * (n_dim - len(a.shape)) + a.shape[1:]
        else:
            t_shape = (1,) * (n_dim - len(a.shape)) + a.shape

        a_reshape = mx.nd.reshape(a, shape=t_shape)
        out = mx.nd.broadcast_to(a_reshape, shape=self.shape)
        self.assign(out_data[0], req[0], out)

    def backward(self, req, out_grad, in_data, out_data, in_grad, aux):
        a_shape = in_data[0].shape
        if self.isSamples:
            grad = mx.nd.reshape(out_grad[0], shape=(a_shape[0], -1,) + a_shape[1:])
            a_grad = mx.nd.sum(grad, axis=1)
        else:
            grad = mx.nd.reshape(out_grad[0], shape=(-1,) + a_shape)
            a_grad = mx.nd.sum(grad, axis=0)
        self.assign(in_grad[0], req[0], a_grad)


@mx.operator.register("broadcast_to_w_samples")
class BroadcastToWithSamplesOpProp(CustomOpProp):
    def __init__(self, **kwargs):
        self.isSamples = kwargs['isSamples'].lower() == 'true'
        self.shape = parse_string_to_tuple(kwargs['shape'])
        super(BroadcastToWithSamplesOpProp, self).__init__(need_top_grad=True)

    def list_arguments(self):
        return ['data']

    def list_outputs(self):
        return ['out']

    def infer_shape(self, in_shapes):
        return in_shapes, (self.shape,), ()

    def create_operator(self, ctx, in_shapes, in_dtypes, **kwargs):
        return BroadcastToWithSamplesOp(isSamples=self.isSamples,
                                        shape=self.shape, **kwargs)


def broadcast_to_w_samples(F, data, shape, isSamples=True):
    if F is mx.nd:
        n_dim = len(shape)
        if isSamples:
            num_samples = max(data.shape[0], shape[0])
            t_shape = (data.shape[0],) + (1,) * (n_dim - len(data.shape)) + data.shape[1:]
            shape = (num_samples,) + shape[1:]
        else:
            t_shape = (1,) * (n_dim - len(data.shape)) + data.shape

        data_reshape = F.reshape(data, shape=t_shape)
        return F.broadcast_to(data_reshape, shape=shape)
    else:
        return F.Custom(data, op_type="broadcast_to_w_samples",
                        isSamples=isSamples, shape=shape)



"""The main module for MXFusion.

Submodules
==========

.. autosummary::
    :toctree: _autosummary

    factor_graph
    model
    posterior
"""

__all__ = ['factor_graph', 'model', 'posterior']

from .model import Model
from .factor_graph import FactorGraph
from .posterior import Posterior



import warnings
import numpy as np
import mxnet as mx
from mxnet import initializer
from mxnet import ndarray
from mxnet.gluon import ParameterDict
from ..components.variables import VariableType, Variable
from ..components import ModelComponent
from ..util.inference import realize_shape
from ..common.config import get_default_device
from ..components.functions.gluon_func_eval import GluonFunctionEvaluation


class InferenceParameters(object):
    """
    The parameters and outcomes of an inference method.

    InferenceParameters is a pool of memory that contains a mapping from uuid to two types of memories (MXNet ParameterDict and Constants).

    :param constants: Specify a list of model variables as constants
    :type constants: {ModelComponent.uuid : mxnet.ndarray}
    :param dtype: data type for internal numberical representation
    :type dtype: {numpy.float64, numpy.float32, 'float64', 'float32'}
    :param context: The MXNet context
    :type context: {mxnet.cpu or mxnet.gpu}
    """
    def __init__(self, constants=None, dtype=None, context=None):
        self.dtype = dtype if dtype is not None else np.float32
        self.mxnet_context = context if context is not None else get_default_device()
        self._constants = {}
        self._var_ties = {}
        if constants is not None:
            constant_uuids = {
                (k.uuid if isinstance(k, ModelComponent) else k): v
                for k, v in constants.items()}
            self._constants.update(constant_uuids)
        self._params = ParameterDict()

    def update_constants(self, constants):
        self.constants.update({
            (k.uuid if isinstance(k, ModelComponent) else k): v
            for k, v in constants.items()})

    def initialize_params(self, graphs, observed_uuid):
        """
        :param graphs: a list of graphs in which the parameters will be optimized.
        :type graphs: a list of FactorGraph
        :param observed_uuid: Parameter Variables that are passed in directly as data, not to be inferred.
        :type observed_uuid: list, set
        """
        if self._params is not None:
            warnings.warn("InferenceParameters has already been initialized.  The existing one will be overwritten.")

        self._params = ParameterDict()
        for g in graphs:
            # load in parameterdict from external gluon blocks.
            for f in g.functions.values():
                if isinstance(f, GluonFunctionEvaluation):
                    self._params.update(
                        f.function_wrapper.collect_internal_parameters())

            for var in g.get_constants():
                self._constants[var.uuid] = var.constant

            excluded = set(self._constants.keys()).union(observed_uuid)
            for var in g.get_parameters(excluded=excluded,
                                        include_inherited=False):
                var_shape = realize_shape(var.shape, self._constants)
                init = initializer.Constant(var.initial_value) if var.initial_value is not None else None

                self._params.get(name=var.uuid, shape=var_shape,
                                 dtype=self.dtype,
                                 allow_deferred_init=True, init=init)

        self._params.initialize(ctx=self.mxnet_context)

    def initialize_with_carryover_params(self, graphs, observed_uuid, var_ties,
                                         carryover_params):
        """
        :param graphs: a list of graphs in which the parameters will be optimized.
        :type graphs: a list of FactorGraph
        :param observed_uuid: Parameter Variables that are passed in directly as data, not to be inferred.
        :type observed_uuid: {UUID : mx.ndarray}
        :param var_ties: A dictionary of variable maps that are tied together and use the MXNet Parameter of the dict value's uuid.
        :type var_ties: { UUID to tie from : UUID to tie to }
        :param carryover_params: list of InferenceParameters containing the outcomes of previous inference algorithms.
        :type carryover_params: [InferenceParameters]
        """
        # TODO: var_ties is discarded at the moment.

        var_uuid = set()
        for g in graphs:
            var_uuid = var_uuid.union(set(g.variables.keys()))

        carryover_pairs = {}
        for carryover in carryover_params:
            for uuid, v in carryover.param_dict.items():
                if uuid in var_uuid:
                    if uuid in carryover_pairs:
                        warnings.warn('The variable with UUID '+uuid+' exists in multiple carryover parameter sets.')
                    carryover_pairs[uuid] = v

        # self._var_ties = var_ties.copy()
        # for g in graphs:
        #     # TODO: check the behavior of var_ties in graph
        #     self._var_ties.update(g.var_ties)
        # for v_uuid in self.constants:
        #     if v_uuid in self._var_ties:
        #         del self._var_ties[v_uuid]

        observed_uuid = set(observed_uuid).union(carryover_pairs.keys())
        self.initialize_params(graphs, observed_uuid)

        # carryover_pairs = {
        #     to_var_uuid: carryover.param_dict[to_var_uuid]
        #     for from_var_uuid, to_var_uuid in self._var_ties.items()
        #     for carryover in carryover_params
        #     if to_var_uuid in carryover.param_dict}
        self._params.update(carryover_pairs)

    @property
    def param_dict(self):
        return self._params

    @property
    def constants(self):
        return self._constants

    @property
    def var_ties(self):
        return self._var_ties

    def __getitem__(self, key, ctx=None):
        if not isinstance(key, Variable):
            raise KeyError("The access key of inference parameter needs to be Variable, but got "+str(type(key))+".")
        pkey = key.inherited_name if key.isInherited else key.uuid
        val = self._params.get(pkey).data(ctx)
        if key.transformation is not None:
            val = key.transformation.transform(val)
        return val

    def __setitem__(self, key, item):
        if not isinstance(key, Variable):
            raise KeyError("The access key of inference parameter needs to be Variable, but get "+str(type(key))+".")

        if key.type == VariableType.PARAMETER:
            if key.transformation is not None:
                item = key.transformation.inverseTransform(item)
            self._params.get(key.uuid).set_data(item)
        elif key.type == VariableType.CONSTANT:
            self._params.get(key.uuid)._value = item

    # Override contains so that it doesn't use the __getitem__ method.
    def __contains__(self, k):
        return k in self.__dict__

    @staticmethod
    def load_parameters(uuid_map=None,
                        parameters_file=None,
                        variable_constants_file=None,
                        mxnet_constants_file=None,
                        context=None, dtype=None,
                        current_params=None):
        """
        Loads back a sest of InferenceParameters from files.
        :param parameters_file: These are the parameters of the previous inference algorithm.  These are in a {uuid: mx.nd.array} mapping.
        :type mxnet_constants_file: file saved down with mx.nd.save(), so a {uuid: mx.nd.array} mapping saved in a binary format.
        :param mxnet_constants_file: These are the constants in mxnet format from the previous inference algorithm. These are in a {uuid: mx.nd.array} mapping.
        :type mxnet_constants_file: file saved down with mx.nd.save(), so a {uuid: mx.nd.array} mapping saved in a binary format.
        :param variable_constants_file: These are the constants in primitive format from the previous inference algorithm.
        :type variable_constants_file: json dict of {uuid: constant_primitive}
        """
        def with_uuid_map(item, uuid_map):
            if uuid_map is not None:
                return uuid_map[item]
            else:
                return item
        ip = InferenceParameters(context=context, dtype=dtype)

        if parameters_file is not None:
            old_params = ndarray.load(parameters_file)
            mapped_params = {with_uuid_map(k, uuid_map): v
                             for k, v in old_params.items()}

            new_paramdict = ParameterDict()
            if current_params is not None:
                new_paramdict.update(current_params)

            # Do this because we need to map the uuids to the new Model
            # before loading them into the ParamDict
            for name, mapped_param in mapped_params.items():
                new_paramdict[name]._load_init(mapped_param, ip.mxnet_context)
            ip._params = new_paramdict

        new_mxnet_constants = {}
        new_variable_constants = {}
        if variable_constants_file is not None:
            import json
            with open(variable_constants_file) as f:
                old_constants = json.load(f)
                new_variable_constants = {with_uuid_map(k, uuid_map): v for k, v in old_constants.items()}
        if mxnet_constants_file is not None:
            new_mxnet_constants = {with_uuid_map(k, uuid_map): v for k, v in ndarray.load(mxnet_constants_file).items()}
        ip._constants = {}
        ip._constants.update(new_variable_constants)
        ip._constants.update(new_mxnet_constants)
        return ip

    def save(self, prefix):
        """
        Saves the parameters and constants down to json files as maps from {uuid : value}, where value is an mx.ndarray for parameters and either primitive number types or mx.ndarray for constants. Saves up to 3 files: prefix+["_params.json", "_variable_constants.json", "_mxnet_constants.json"]

        :param prefix: The directory and any appending tag for the files to save this Inference as.
        :type prefix: str , ex. "../saved_inferences/experiment_1"
        """
        param_file = prefix + "_params.json"
        variable_constants_file = prefix + "_variable_constants.json"
        mxnet_constants_file = prefix + "_mxnet_constants.json"
        to_save = {key: value._reduce() for key, value in self._params.items()}
        ndarray.save(param_file, to_save)

        mxnet_constants = {uuid: value
                           for uuid, value in self._constants.items()
                           if isinstance(value, mx.ndarray.ndarray.NDArray)}
        ndarray.save(mxnet_constants_file, mxnet_constants)

        variable_constants = {uuid: value
                              for uuid, value in self._constants.items()
                              if uuid not in mxnet_constants}
        import json
        with open(variable_constants_file, 'w') as f:
            json.dump(variable_constants, f, ensure_ascii=False)



from .gluon_func_eval import GluonFunctionEvaluation
from ..variables import Variable, VariableType
from mxnet.gluon import ParameterDict, Block, HybridBlock
from ...common.exceptions import ModelSpecificationError


class MXFusionGluonFunction(object):
    """
    The wrapper of a MXNet Gluon block in MXFusion. It automatically fetches all the Gluon parameters in its ParameterDict. When this function
    wrapper is called in Model definition, it returns a factor corresponding to the function evaluation.

    :param block: The MXNet Gluon block to be wrapped.
    :type block: mxnet.gluon.Blockk or mxnet.gluon.HybridBlock
    :param num_outputs: The number of output variables of the Gluon block.
    :type num_outputs: int
    :param dtype: the data type of float point numbers used in the Gluon block.
    :type dtype: numpy.float32 or numpy.float64
    :param broadcastable: Whether the function supports broadcasting with the additional dimension for samples.
    :type broadcastable: boolean
    """

    def __init__(self, block, num_outputs, dtype=None, broadcastable=False):

        if not isinstance(block, (Block, HybridBlock)):
            raise ModelSpecificationError('The block argument must be of type Block or HybridBlock from MXNet Gluon.')

        super(MXFusionGluonFunction, self).__init__()
        self.broadcastable = broadcastable
        self.block = block
        self._name = block.name
        self.num_outputs = num_outputs
        self._gluon_parameters = self._create_variables_from_gluon_block(block)

    def __call__(self, *args, **kwargs):
        """
        Bind the wrapped Gluon block with the input arguments. This is called during model specification.

        :param args: the positional arguments that the Gluon block takes.
        :type args: [Variable]
        :param kwargs: Internal block variables the user would like to overwrite with Random or Function Variables.
        :type kwargs: {name : Variable}
        :returns: The output variables of the FunctionEvaluation with the specified inputs.
        :rtype: A tuple of output Variables if >1 or a single output Variable if 1
        """

        # Copy the block variables so that the user can overwrite them with Random or Function Variables without
        # disturbing other FunctionEvaluations.
        # TODO: support binding (tying) a gluon parameter with a normal model parameter
        block_variables = self._gluon_parameters.copy()
        block_variables.update(kwargs)
        block_variables = [(k, v) for k, v in block_variables.items()]

        input_variables = [(self.block.name + "_input_" + str(i), variable) for i, variable in enumerate(args)]

        outputs = []
        for i in range(self.num_outputs):
            output = Variable()
            outputs.append((self.name + "_output_" + str(i), output))

        fe = GluonFunctionEvaluation(
            self, input_variables, block_variables, outputs,
            broadcastable=self.broadcastable)

        return tuple([v for _, v in fe.outputs]) if len(fe.outputs) > 1 else fe.outputs[0][1]

    def _create_variables_from_gluon_block(self, block):
        """
        Create a list of Parameter type variables from a MXNet Gluon block's parameters.
        One variable per parameter.

        :param block: The Block to create variables from.
        :rtype: {'block_variable.name' : block_variable}
        """
        params = block.collect_params()
        vs = {}
        for param in params.values():
            v = Variable(isInherited=True, shape=param.shape)
            v.inherited_name = param.name
            vs[v.inherited_name] = v
        return vs

    def collect_internal_parameters(self):
        """
        Return the parameters of the MXNet Gluon block that have *not* been set a prior distribution.

        :returns: the parameters of the MXNet Gluon block without a prior distribution.
        :rtype: MXNet.gluon.ParameterDict
        """
        params = ParameterDict()
        gluon_params = self.block.collect_params()
        params.update({var_name: gluon_params[var_name] for var_name, var in self._gluon_parameters.items() if var.type == VariableType.PARAMETER})
        return params

    def collect_params(self):
        """
        Return a variable set / dict. Used for the function.collect_params.set_prior() functionality.
        """
        # TODO: implement VariableSet
        raise NotImplementedError

    @property
    def name(self):
        return self._name



"""This module contains inference related methods and classes.

Submodules
==========

.. autosummary::
    :toctree: _autosummary

    batch_loop
    forward_sampling
    grad_based_inference
    grad_loop
    inference_alg
    inference_parameters
    inference
    map
    meanfield
    minibatch_loop
    variational
"""


from .map import MAP
from .batch_loop import BatchInferenceLoop
from .inference import Inference
from .minibatch_loop import MinibatchInferenceLoop
from .meanfield import create_Gaussian_meanfield
from .forward_sampling import ForwardSampling, VariationalPosteriorForwardSampling
from .grad_based_inference import GradBasedInference
from .variational import StochasticVariationalInference
from .inference_parameters import InferenceParameters



import re
import mxnet as mx
import numpy as np
from mxnet.ndarray.ndarray import NDArray
from mxnet.symbol.symbol import Symbol


def slice_axis(F, array, axis, indices):
    """
    Slice an array along a given axis by a list of indices.

    :param array: the array to be sliced.
    :type array: MXNet Array
    :param axis: the axis to be sliced.
    :type axis: int
    :param indices: the indices used in slicing
    :type indices: list or MXNet Array
    """
    assert F == mx.nd, "The slice_axis helper funcion only works on imperative mode, because fancy indexing only exists in NDArray API."
    if isinstance(indices, (list, tuple)):
        num_indices = len(indices)
    elif isinstance(indices, (NDArray, Symbol)):
        num_indices = indices.shape[0]
    if axis < 0:
        axis = len(array.shape) + axis
    if axis == 0:
        target_shape = (num_indices,) + array.shape[:1]
        array = F.reshape(array, shape=(array.shape[0], -1))[indices, :]
    elif axis == len(array.shape) - 1:
        target_shape = array.shape[:-1] + (num_indices,)
        array = F.reshape(array, shape=(-1, array.shape[-1]))[:, indices]
    else:
        target_shape = array.shape[:axis] + (num_indices,) + \
            array.shape[axis + 1:]

        def multiply(l):
            r = 1
            for i in l:
                r *= i
            return r

        size_before = multiply(array.shape[:axis])
        size_after = multiply(array.shape[axis + 1:])

        array = F.reshape(array, shape=(size_before, array.shape[axis],
                                        size_after))[:, indices, :]
    return F.reshape(array, shape=target_shape)


def rename_duplicate_names(names):
    """
    Given a list of names, rename the duplicated names by appending an integer at the end. This function returns a list of tuples each of
    which contains an index of the duplicate name and a new name.
    For example,
    ['a', 'b', 'a', 'a'] -> [(2, 'a0'), (3, 'a0')]
    ['a', 'b'] -> []

    :param names: the list of names
    :type names: [str]
    :returns: a list of tuples each of which contains an index of the duplicate name and a new name.
    :rtypes: [tuple]
    """
    all_names = set(names)
    if len(all_names) == len(names):
        return []
    cur_names = set()
    prog = re.compile(r'^(.*)(\d+)$')
    renames = []
    for i, n in enumerate(names):
        if n in cur_names:
            res = prog.match(n)
            if res is None or len(res.groups())==0:
                prefix = n
                count = 0
            else:
                prefix = res.groups()[0]
                count = int(res.groups()[1])+1
            while prefix + str(count) in all_names:
                count += 1
            renames.append((i, prefix + str(count)))
            all_names.add(prefix + str(count))
        else:
            cur_names.add(n)
    return renames


def parse_string_to_tuple(s):
    res = re.match(r'[\(\[](.*)[\)\]]', s)
    if res is None or len(res.groups())==0:
        raise Exception
    s_mid = res.groups()[0]
    return tuple(int(i) for i in s_mid.split(','))


def add_mxnet_params(block, var, grad_req='write'):
    """
    This function adds MXFusion variables to the MXNet Gluon block's parameters for use in computations via the params variable or kwargs
    parameter in hybrid_forward()

    :param block: MXNet Gluon Block to add the parameters into
    :param var: array of MXFusion variables to add to the block
    :param grad_req: gradient required for these variables?
    :return:
    """
    from ..components.variables import Variable, VariableType
    param_dict = block.params
    with block.name_scope():
        for v in var:
            if isinstance(v, Variable) and v.type == VariableType.CONSTANT:
                if v.is_scalar:
                    init = mx.initializer.Constant(v.val)
                    init.dumps()
                    parameter = param_dict.get(v.name, shape=v.shape, init=init, grad_req='null')
                    block.__setattr__(v.name, parameter)
            else:
                block.__setattr__(v.name, param_dict.get(v.name, shape=v.shape, init=mx.initializer.One(),
                grad_req=grad_req))


def create_variables_with_names(names):
    """
    Utility function to create a dictionary of Variables from a list of names.

    :param: names list of names
    """
    from ..components.variables import Variable
    return {n: Variable(n) for n in names}


def create_constant_from_values(var):
    """
    Utility function to createa a constant variable from a raw value.

    :param: var the value of the constant
    """
    from ..components.variables import Constant

    if isinstance(var, (int, float, np.ndarray, mx.nd.NDArray)):
        return Constant(var)
    elif isinstance(var, (list, tuple)):
        return Constant(mx.nd.array(var))
    return var



import mxnet as mx
import numpy as np
import mxfusion as mf
import mxnet.gluon.nn as nn
from mxfusion.components.variables.var_trans import PositiveTransformation
from mxnet.gluon import HybridBlock
from ..components.distributions.random_gen import RandomGenerator
from ..components.variables import add_sample_dimension

def prepare_mxnet_array(array, is_sampled_array, dtype):
    a_mx = mx.nd.array(array, dtype=dtype)
    if not is_sampled_array:
        a_mx = add_sample_dimension(mx.nd, a_mx)
    return a_mx


def numpy_array_reshape(var, isSamples, n_dim):
    """
    Reshape a numpy to a give dimensionality by adding dimensions with size one in front (broadcasting). If *isSamples* is true, keep the first dimension.

    :param var: the variable to be reshaped.
    :type var: numpy.ndarray
    :param isSamples: whether the variable is a sampled variable with the first dimension being the number of samples.
    :type isSamples: boolean
    :param n_dim: the dimensionality that the variable is reshaped to.
    :type n_dim: int
    :returns: the reshaped numpy array
    :rtype: numpy.ndarray
    """
    if len(var.shape) < n_dim:
        if isSamples:
            var_reshape = var.reshape(
                *((var.shape[0],) + (1,) * (n_dim - len(var.shape)) +
                  var.shape[1:]))
        else:
            var_reshape = var.reshape(*((1,) * (n_dim - len(var.shape)) +
                                      var.shape))
    else:
        var_reshape = var
    return var_reshape


class MockMXNetRandomGenerator(RandomGenerator):
    """
    The MXNet pseudo-random number generator.
    """

    def __init__(self, samples):
        self._samples = samples

    def sample_normal(self, loc=0, scale=1, shape=None, dtype=None, out=None,
                      ctx=None, F=None):
        if shape is None:
            shape = (1,)
        res = mx.nd.reshape(self._samples[:np.prod(shape)], shape=shape)
        if out is not None:
            out[:] = res
        return res

    def sample_multinomial(self, data, shape=None, get_prob=True, dtype='int32',
                           F=None):
        return mx.nd.reshape(self._samples[:np.prod(data.shape[:-1])], shape=data.shape[:-1])


def make_net():
    D=100
    net = nn.HybridSequential(prefix='testnet0_')
    with net.name_scope():
        net.add(nn.Dense(D, activation="tanh"))
        net.add(nn.Dense(D, activation="tanh"))
        net.add(nn.Dense(1, flatten=True))
    net.initialize(mx.init.Xavier(magnitude=3))
    return net


def make_basic_model(finalize=True, verbose=True):
    net = make_net()
    m = mf.models.Model(verbose=verbose)
    m.N = mf.components.Variable()
    m.x = mf.components.Variable(shape=(m.N,))
    m.v = mf.components.Variable(shape=(1,), transformation=PositiveTransformation())
    m.r = mf.components.Variable(shape=(m.N,))
    m.y = mf.components.distributions.Normal.define_variable(mean=m.r, variance=m.v, shape=(m.N,))
    return m

def make_bnn_model(finalize=True, verbose=True):
    net = make_net()
    m = mf.models.Model(verbose=verbose)
    m.N = mf.components.Variable()
    m.f = mf.components.MXFusionGluonFunction(net, num_outputs=1)
    m.x = mf.components.Variable(shape=(m.N,))
    m.v = mf.components.Variable(shape=(1,), transformation=PositiveTransformation())
    m.prior_variance = mf.components.Variable(shape=(1,), transformation=PositiveTransformation())
    m.r = m.f(m.x)
    for _, v in m.r.factor.block_variables:
        v.set_prior(mf.components.distributions.Normal(mean=mx.nd.array([0]), variance=m.prior_variance))
    m.y = mf.components.distributions.Normal.define_variable(mean=m.r, variance=m.v, shape=(m.N,))
    return m


class DotProduct(HybridBlock):
    def hybrid_forward(self, F, x, *args, **kwargs):
        return F.dot(x, args[0])

class TestBlock(mx.gluon.HybridBlock):
    """
    Block with standard functions for initializing and running an MXNet Gluon block for unit testing.
    """

    def __init__(self):
        """
        Constructor.
        """

        super(TestBlock, self).__init__()
        with self.name_scope():
            self.var1 = self.params.get('var1', shape=(2,))
            self.var2 = self.params.get('var2', shape=(2,))

    def hybrid_forward(self, F, x, var1, var2):
        """
        Simple test function to enable MXFusion tests that use a Gluon block.
        :param F: MXNet computation type <mx.sym, mx.nd>
        :param x: MXNet dummy variable
        :param var1: MXNet dummy variable
        :param var2: MXNet dummy variable
        """

        return x + var1 + var2



from .kernel import NativeKernel
from ....variables import Variable
from ....variables import PositiveTransformation
from .....util.customop import broadcast_to_w_samples


class Linear(NativeKernel):
    """
    Linear kernel

    .. math::
       k(x,y) = \\sum_{i=1}^{\\text{input_dim}} \\sigma^2_i x_iy_i

    :param input_dim: the number of dimensions of the kernel. (The total number of active dimensions) .
    :type input_dim: int
    :param ARD: a binary switch for Automatic Relevance Determination (ARD). If true, the squared distance is divided by a lengthscale for individual
        dimensions.
    :type ARD: boolean
    :param variances: the initial value for the variances parameter, which scales the input dimensions.
    :type variances: float or MXNet NDArray
    :param name: the name of the kernel. The name is used to access kernel parameters.
    :type name: str
    :param active_dims: The dimensions of the inputs that are taken for the covariance matrix computation. (default: None, taking all the dimensions).
    :type active_dims: [int] or None
    :param dtype: the data type for float point numbers.
    :type dtype: numpy.float32 or numpy.float64
    :param ctx: the mxnet context (default: None/current context).
    :type ctx: None or mxnet.cpu or mxnet.gpu
    """
    def __init__(self, input_dim, ARD=False, variances=1., name='linear',
                 active_dims=None, dtype=None, ctx=None):
        super(Linear, self).__init__(input_dim=input_dim, name=name,
                                     active_dims=active_dims, dtype=dtype,
                                     ctx=ctx)
        self.ARD = ARD
        if not isinstance(variances, Variable):
            variances = Variable(shape=(input_dim if ARD else 1,),
                                 transformation=PositiveTransformation(),
                                 initial_value=variances)
        self.variances = variances

    def _compute_K(self, F, X, variances, X2=None):
        """
        The internal interface for the actual covariance matrix computation.

        :param F: MXNet computation type <mx.sym, mx.nd>.
        :param X: the first set of inputs to the kernel.
        :type X: MXNet NDArray or MXNet Symbol
        :param X2: (optional) the second set of arguments to the kernel. If X2 is None, this computes a square covariance matrix of X. In other words,
            X2 is internally treated as X.
        :type X2: MXNet NDArray or MXNet Symbol
        :param variances: the variances parameter, which scales the input dimensions.
        :type variances: MXNet NDArray or MXNet Symbol
        :param lengthscale: the lengthscale parameter.
        :type lengthscale: MXNet NDArray or MXNet Symbol
        :return: The covariance matrix.
        :rtype: MXNet NDArray or MXNet Symbol
        """
        if self.ARD:
            var_sqrt = F.sqrt(variances)
            if X2 is None:
                xsc = X * broadcast_to_w_samples(F, var_sqrt, X.shape)
                return F.linalg.syrk(xsc)
            else:
                xsc = X * broadcast_to_w_samples(F, var_sqrt, X.shape)
                x2sc = X2 * broadcast_to_w_samples(F, var_sqrt, X2.shape)
                return F.linalg.gemm2(xsc, x2sc, False, True)
        else:
            if X2 is None:
                A = F.linalg.syrk(X)
            else:
                A = F.linalg.gemm2(X, X2, False, True)
            return A * broadcast_to_w_samples(F, variances, A.shape)

    def _compute_Kdiag(self, F, X, variances):
        """
        The internal interface for the actual computation for the diagonal of the covariance matrix.

        :param F: MXNet computation type <mx.sym, mx.nd>.
        :param X: the first set of inputs to the kernel.
        :type X: MXNet NDArray or MXNet Symbol
        :param variances: the variances parameter, which scales the input dimensions.
        :type variances: MXNet NDArray or MXNet Symbol
        :return: The covariance matrix.
        :rtype: MXNet NDArray or MXNet Symbol
        """
        X2 = F.square(X)
        return F.sum(X2 * broadcast_to_w_samples(F, variances, X2.shape),
                     axis=-1)



from future.utils import raise_from
from uuid import uuid4
import warnings
import networkx as nx
from ..components import Distribution, Factor, ModelComponent, Module, Variable, VariableType
from ..common.exceptions import ModelSpecificationError, InferenceError
from ..components.functions import FunctionEvaluation
from ..components.variables.runtime_variable import expectation


class FactorGraph(object):
    """
    A graph defining how Factor objects relate to one another.

    The two primary functionalities of this class are:

     * ``compute_log_prob`` which computes the log probability of some variables in the graph and
     * ``draw_samples`` which draws samples for some target variables from the graph
    """

    def __init__(self, name, verbose=False):
        """
        Initializes the FactorGraph with a UUID and structures to hold components.
        """
        self.name = name
        self._uuid = str(uuid4())
        self._var_ties = {}

        self._components_graph = nx.DiGraph()
        self._verbose = verbose

    def __repr__(self):
        """
        Return a string summary of this object
        """
        out_str = ''
        for f in self.ordered_factors:
            if isinstance(f, FunctionEvaluation):
                out_str += ', '.join([str(v) for _, v in f.outputs])+' = '+str(f)+'\n'
            elif isinstance(f, Distribution):
                out_str += ', '.join([str(v) for _, v in f.outputs])+' ~ '+str(f)+'\n'
        return out_str[:-1]

    def __getitem__(self, key):
        return self.components[key]

    def __contains__(self, key):
        return key in self.components

    def __setattr__(self, name, value):
        """
        Called whenever an attribute is attached to a FactorGraph object.
        This method adds the attribute's value to it's internal graph representation
        if it's an object derived from the ModelComponent class.

        :param name:  The attribute name.
        :type name: str
        :param value: The value being assigned to the attribute.
        :type value: Anything, but if it is type ModelComponent it is added to some internal data structures.
        """
        if isinstance(value, ModelComponent):
            if self._verbose:
                print("Variable {} ({}) {} ({})".format(name, value.uuid, value.type, value.shape))
            if value.name is not None:
                warnings.warn("The value {} has already been assigned in the model.".format(str(value)))
            value.name = name
            value.graph = self.components_graph
        super(FactorGraph, self).__setattr__(name, value)

    @property
    def components_graph(self):
        """
        Return the Graph object of the component

        :returns: dict of ModelComponents
        :rtype: { UUID : ModelComponent }
        """

        return self._components_graph

    @property
    def components(self):
        """
        Return all the ModelComponents held in the FactorGraph.

        :returns: dict of ModelComponents
        :rtype: { UUID : ModelComponent }
        """

        return {node.uuid: node for node in self.components_graph.nodes()}

    @property
    def distributions(self):
        """
        Return the distributions held in the FactorGraph.

        :returns: dict of Distributions
        :rtype: { UUID : Distribution }
        """

        return {node.uuid: node for node in self.components_graph.nodes() if isinstance(node, Distribution)}

    @property
    def functions(self):
        """
        Return the functions held in the FactorGraph.

        :returns: dict of Functions
        :rtype: { UUID : Function }
        """

        return {node.uuid: node for node in self.components_graph.nodes() if isinstance(node, FunctionEvaluation)}

    @property
    def modules(self):
        """
        Return the modules held in the FactorGraph.

        :returns: dict of Modules
        :rtype: { UUID : Module }
        """

        return {node.uuid: node for node in self.components_graph.nodes() if isinstance(node, Module)}

    @property
    def variables(self):
        """
        Return the variables held in the FactorGraph.

        :returns: dict of Variables
        :rtype: { UUID : Variable }
        """

        return {node.uuid: node for node in self.components_graph.nodes() if isinstance(node, Variable)}

    @property
    def ordered_factors(self):
        """
        Return a sorted list of Factors in the graph.

        :rtype: A topologically sorted list of Factors in the graph.
        """
        return [node for node in nx.topological_sort(self.components_graph)
                                 if isinstance(node, Factor)]

    @property
    def roots(self):
        """
        Return all root notes in the graph.
        """
        return [node for node, degree in self.components_graph.in_degree() if degree == 0]

    @property
    def leaves(self):
        """
        Return all leaf nodes in the graph.
        """
        return [node for node, degree in self.components_graph.out_degree() if degree == 0]

    @property
    def var_ties(self):
        """
        Return a mapping of Variables in the FactorGraph that are tied together (i.e. the same / aliases of each other.)

        :returns: dict of UUIDs of the variables to tie
        :rtype: { uuid_of_var_to_tie : uuid_of_var_to_tie_to }
        """

        return self._var_ties

    def compute_log_prob(self, F, targets, conditionals=None, constants=None):
        """
        Compute the logarithm of the probability/probability density of a set of random variables in the factor graph. The set of random
        variables are specified in the "target" argument and any necessary conditional variables are specified in the "conditionals" argument.
        Any relevant constants are specified in the "constants" argument.

        :param F: the MXNet computation mode (``mxnet.symbol`` or ``mxnet.ndarray``).
        :param targets: Variables to compute the log probability of.
        :type targets: {uuid : RTVariable}
        :param conditionals: variables to condition the probabilities on.
        :type conditionals: {uuid : RTVariable}
        :param constants: the constants that may be used in computation.
        :type constants: {UUID: float or mxnet NDArray or mxnet Symbol}
        :returns: the sum of the log probability of all the target variables.
        :rtype: RTVariable
        """
        constants = {} if constants is None else constants

        variables = targets.copy()
        variables.update(constants)
        if conditionals is not None:
            variables.update(conditionals)

        logL = 0.
        for f in self.ordered_factors:
            if isinstance(f, FunctionEvaluation):
                outcome = f.eval(F=F, variables=variables,
                                 always_return_tuple=True)
                outcome_uuid = [v.uuid for _, v in f.outputs]
                for v, uuid in zip(outcome, outcome_uuid):
                    if uuid in variables:
                        warnings.warn('Function evaluation in FactorGraph.compute_log_prob_RT: the outcome variable '+str(uuid)+' of the function evaluation '+str(f)+' has already existed in the variable set.')
                    variables[uuid] = v
            elif isinstance(f, Distribution):
                if f.random_variable.uuid in targets:
                    logL = logL + F.sum(expectation(F, f.log_pdf(
                        F=F, variables=variables)))
            elif isinstance(f, Module):
                raise NotImplementedError("Modules aren't implemented yet!")
            else:
                raise ModelSpecificationError("There is an object in the factor graph that isn't a factor." + "That shouldn't happen.")
        return logL

    def draw_samples(self, F, num_samples=1, targets=None, conditionals=None,
                     constants=None):
        """
        Draw samples from the target variables of the Factor Graph. If the ``targets`` argument is None, draw samples from all the variables
        that are *not* in the ``conditionals`` argument.

        :param F: the MXNet computation mode (``mxnet.symbol`` or ``mxnet.ndarray``).
        :param num_samples: The number of samples to draw for the target variables.
        :type num_samples: int
        :param targets: a list of Variables to draw samples from.
        :type targets: [UUID]
        :param conditionals: Variables to condition the samples on.
        :type conditionals: {UUID : RTVariable}
        :param constants: the constants that may be used in computation.
        :type constants: {UUID: float or mxnet NDArray or mxnet Symbol}
        :returns: the samples of the target variables.
        :rtype: {UUID : RTVariable}
        """
        constants = {} if constants is None else constants
        variables = {}
        variables.update(constants)
        if conditionals is not None:
            variables.update(conditionals)

        samples = {}
        for f in self.ordered_factors:
            if isinstance(f, FunctionEvaluation):
                outcome = f.eval(F=F, variables=variables,
                                 always_return_tuple=True)
                outcome_uuid = [v.uuid for _, v in f.outputs]
                for v, uuid in zip(outcome, outcome_uuid):
                    if uuid in variables:
                        warnings.warn('Function evaluation in FactorGraph.draw_samples_RT: the outcome of the function evaluation '+str(f)+' has already existed in the variable set.')
                    variables[uuid] = v
                    samples[uuid] = v
            elif isinstance(f, Distribution):
                known = [v in variables for _, v in f.outputs]
                if all(known):
                    continue
                elif any(known):
                    raise InferenceError("Part of the outputs of the distribution " + f.__class__.__name__ + " has been observed!")
                outcome = f.draw_samples(
                    F=F, num_samples=num_samples, variables=variables,
                    always_return_tuple=True)
                outcome_uuid = [v.uuid for _, v in f.outputs]
                for v, uuid in zip(outcome, outcome_uuid):
                    variables[uuid] = v
                    samples[uuid] = v
            elif isinstance(f, Module):
                raise NotImplementedError("Modules aren't implemented yet!")
            else:
                raise ModelSpecificationError("There is an object in the factor graph that isn't a factor." + "That shouldn't happen.")
        if targets:
            samples = {uuid: samples[uuid] for uuid in targets}
        return samples

    def remove_component(self, component):
        """
        Removes the specified component from the factor graph.

        :param component: The component to remove.
        :type component: ModelComponent
        """
        if not isinstance(component, ModelComponent):
                raise ModelSpecificationError(
                    "Attempted to remove an object that isn't a ModelComponent.")

        try:
            self.components_graph.remove_node(component)  # implicitly removes edges
        except Exception as e:
            raise_from(ModelSpecificationError("Attempted to remove a node that isn't in the graph"), e)

        if component.name is not None:
            try:
                self.__delattr__(component.name)
            except Exception as e:
                pass

        component.graph = None

    def _replicate_class(self, **kwargs):
        """
        Returns a new instance of the derived FactorGraph's class.
        """
        return FactorGraph(**kwargs)

    def remove_subgraph(self, node):
        """
        Removes a node and its parent graph recursively.
        """
        if isinstance(node, Variable):
            self.remove_component(node)
            if node.factor is not None:
                self.remove_subgraph(node.factor)
        elif isinstance(node, Factor):
            self.remove_component(node)
            for _, v in node.inputs:
                self.remove_subgraph(v)
        del node

    def replace_subgraph(self, target_variable, new_subgraph):
        """
        Replaces the target_variable with the new_subgraph.

        TODO If the factor of target_variable or new_subgraph as multiple outputs, this will fail.

        :param target_variable: The variable that will be replaced by the new_subgraph.
        :type target_variable: Variable
        :param new_subgraph: We assume this is a Variable right now (aka the graph ends in a Variable).
        :type new_subgraph: Variable
        """
        new_factor = new_subgraph.factor
        new_factor.successors = []

        old_predecessors = target_variable.predecessors
        target_variable.predecessors = []
        for _, p in old_predecessors:
            self.remove_subgraph(p)

        target_variable.assign_factor(new_factor)

    def extract_distribution_of(self, variable):
        """
        Extracts the distribution of the variable passed in, returning a replicated copy of the passed in variable with only its parent
        subgraph attached (also replicated).

        :param variable: The variable to extract the distribution from.
        :type variable: Variable
        :returns: a replicated copy of the passed in variable.
        :rtype: Variable
        """
        def extract_distribution_function(component):
            if isinstance(component, Factor):
                predecessor_direction = 'recursive'
                successor_direction = 'one_level'
                return predecessor_direction, successor_direction
            else:
                predecessor_direction = 'recursive'
                successor_direction = None
                return predecessor_direction, successor_direction
        return variable.replicate(replication_function=extract_distribution_function)

    def clone(self, leaves=None):
        """
        Clones a model, maintaining the same functionality and topology. Replicates all of its ModelComponents with new UUIDs.
        Starts upward from the leaves and copies everything in the graph recursively.

        :param leaves: If None, use the leaves in this model, otherwise use the provided leaves.
        :returns: A tuple of (cloned_model, variable_map)
        """

        new_model = self._replicate_class(name=self.name, verbose=self._verbose)
        var_map = {}  # from old model to new model

        leaves = self.leaves if leaves is None else leaves
        for v in leaves:
            if v.name is not None:
                new_leaf = v.replicate(var_map=var_map,
                           replication_function=lambda x: ('recursive', 'recursive'))
                setattr(new_model, v.name, new_leaf)
            else:
                v.graph = new_model.graph
        for v in self.variables.values():
            if v.name is not None:
                setattr(new_model, v.name, new_model[v.uuid])
        return new_model, var_map

    def get_parameters(self, excluded=None, include_inherited=False):
        """
        Get all the parameters not in the excluded list.

        :param excluded: a list of variables to be excluded.
        :type excluded: set(UUID) or [UUID]
        :param include_inherited: whether inherited variables are included.
        :type include_inherited: boolean
        :returns: the list of contant variables.
        :rtype: [Variable]
        """
        if include_inherited:
            return [v for v in self.variables.values() if (v.type == VariableType.PARAMETER and v.uuid not in excluded)]
        else:
            return [v for v in self.variables.values() if (v.type == VariableType.PARAMETER and v.uuid not in excluded and not v.isInherited)]

    def get_constants(self):
        """
        Get all the contants in the factor graph.

        :returns: the list of constant variables.
        :rtype: [Variable]
        """
        return [v for v in self.variables.values() if v.type == VariableType.CONSTANT]

    @staticmethod
    def reconcile_graphs(current_graphs, primary_previous_graph, secondary_previous_graphs=None, primary_current_graph=None):
        """
        Reconciles two sets of graphs, matching the model components in the previous graph to the current graph.
        This is primarily used when loading back a graph from a file and matching it to an existing in-memory graph in order to load the previous
        graph's paramters correctly.

        :param current_graphs: A list of the graphs we are reconciling a loaded factor graph against. This must be a fully built set of graphs
            generated through the model definition process.
        :param primary_previous_graph: A graph which may have been loaded in from a file and be partially specified, or could be a full graph
            built through model definition.
        :param secondary_previous_graphs: A list of secondary graphs (e.g. posteriors) that share components with the primary_previous_graph.
        :param primary_current_graph: Optional parameter to specify the primary_current_graph, otherwise it is taken to be the model in the
            current_graphs (which should be unique).

        :rtype: {previous ModelComponent : current ModelComponent}
        """
        from .model import Model
        component_map = {}
        current_level = {}
        current_graph = primary_current_graph if primary_current_graph is not None else [graph for graph in current_graphs if isinstance(graph, Model)][0]
        secondary_current_graphs = [graph for graph in current_graphs
                                    if not isinstance(graph, Model)]

        # Map over the named components.
        for c in primary_previous_graph.components.values():
            if c.name:
                current_c = getattr(current_graph, c.name)
                component_map[c.uuid] = current_c.uuid
                current_level[c.uuid] = current_c.uuid

        # Reconcile the primary graph
        FactorGraph._reconcile_graph(current_level, component_map,
                                     current_graph, primary_previous_graph)
        # Reconcile the other graphs
        if not(secondary_current_graphs is None or
               secondary_previous_graphs is None):
            for cg, pg in zip(secondary_current_graphs,
                              secondary_previous_graphs):
                current_level = {pc: cc for pc, cc in component_map.items()
                                 if pc in pg.components.keys()}
                FactorGraph._reconcile_graph(
                    current_level, component_map, cg, pg)

        # Resolve the remaining ambiguities here.
        # if len(component_map) < set([graph.components for graph in previous_graphs])): # TODO the components of all the graphs not just the primary
        #     pass
        return component_map

    @staticmethod
    def _reconcile_graph(current_level, component_map, current_graph, previous_graph):
        """
        Traverses the components in current_level of the current_graph/previous_graph, matching components where possible and generating
        new calls to _reconcile_graph where the graph is still incompletely traversed. This method makes no attempt to resolve ambiguities
        in naming between the graphs and request the user to more completely specify names in their graph if such an ambiguity exists. Such
        naming can be [more] completely specified by attaching names to each leaf node in the original graph.

        :param current_level: A list of items to traverse the graph upwards from.
        :type current_level: [previous ModelComponents]
        :param component_map: The current mapping from the previous graph's MCs to the current_graph's MCs. This is used and modified during reconciliation.
        :type component_map: {previous_graph ModelComponent : current_graph ModelComponent}
        :param current_graph: The current graph to match components against.
        :type current_graph: FactorGraph
        :param previous_graph: The previous graph to match components from.
        :type previous_graph: FactorGraph
        """

        def reconcile_direction(direction, c, current_c, new_level, component_map):
            if direction == 'predecessor':
                previous_neighbors = c.predecessors
                current_neighbors = current_c.predecessors
            elif direction == 'successor':
                previous_neighbors = c.successors
                current_neighbors = current_c.successors
            names = list(map(lambda x: x[0], previous_neighbors))
            duplicate_names = set([x for x in names if names.count(x) > 1])
            for edge_name, node in previous_neighbors:
                if node.uuid not in component_map:
                    if edge_name in duplicate_names:
                        # TODO if all the other parts of the ambiguity are resolved, we have the answer still. Otherwise throw an exception
                        raise Exception("Multiple edges connecting unnamed nodes have the same name, this isn't supported currently.") # TODO Support the ambiguities :)
                    current_node = [item for name, item in current_neighbors if edge_name == name][0]
                    component_map[node.uuid] = current_node.uuid
                    new_level[node.uuid] = current_node.uuid

        new_level = {}
        for c, current_c in current_level.items():
            reconcile_direction('predecessor', previous_graph[c], current_graph[current_c], new_level, component_map)
            """
            TODO Reconciling in both directions currently breaks the reconciliation process and can cause multiple previous_uuid's to map to the same current_uuid. It's unclear why that happens.
            This shouldn't be necessary until we implement multi-output Factors though (and even then, only if not all the outputs are in a named chain).
            """
            # reconcile_direction('successor', previous_graph[c], current_graph[current_c], new_level, component_map)
        if len(new_level) > 0:
            return FactorGraph._reconcile_graph(new_level, component_map, current_graph, previous_graph)

    def load_graph(self, graph_file):
        """
        Method to load back in a graph. The graph file should be saved down using the save method, and is a JSON representation of the graph
        generated by the [networkx](https://networkx.github.io) library.

        :param graph_file: The file containing the primary model to load back for this inference algorithm.
        :type graph_file: str of filename
        """
        import json
        from ..util.graph_serialization import ModelComponentDecoder
        with open(graph_file) as f:
            json_graph = json.load(f, cls=ModelComponentDecoder)

        components_graph = nx.readwrite.json_graph.node_link_graph(
            json_graph, directed=True)
        components = {node.uuid: node for node in components_graph.nodes()}
        for node in components_graph.nodes():
            node._parent_graph = components_graph
            node.attributes = [components[a] for a in node.attributes]
        self._components_graph = components_graph
        for node in self._components_graph.nodes():
            if node.name is not None:
                self.__setattr__(node.name, node)
        return self

    def save(self, graph_file):
        """
        Method to save this graph down into a file. The graph file will be saved down as a JSON representation of the graph generated by the
        [networkx](https://networkx.github.io) library.

        :param graph_file: The file containing the primary model to load back for this inference algorithm.
        :type graph_file: str of filename
        """
        json_graph = nx.readwrite.json_graph.node_link_data(self._components_graph)
        import json
        from ..util.graph_serialization import ModelComponentEncoder
        with open(graph_file, 'w') as f:
            json.dump(json_graph, f, ensure_ascii=False, cls=ModelComponentEncoder)



try:
    import configparser
except:
    import ConfigParser as configparser
import os
import mxnet

config = configparser.ConfigParser()
config.read(os.path.join(os.path.abspath(os.path.dirname(__file__)),
            'config.cfg'))

def get_default_dtype():
    return config['defaults']['mxnet_dtype']


def get_default_MXNet_mode():
    """
    Return the default MXNet mode

    :returns: an MXNet ndarray or symbol, indicating the default mode.
    :rtype: :class:`mxnet.symbol` or :class:`mxnet.ndarray`
    """
    if config['defaults']['mxnet_mode'] == 'mxnet_ndarray':
        return mxnet.ndarray
    elif config['defaults']['mxnet_mode'] == 'mxnet_symbol':
        return mxnet.symbol


def get_default_device():
    """
    Return the default MXNet device

    :returns: an MXNet cpu or gpu, indicating the default device.
    """
    if config['defaults']['mxnet_device'] == 'cpu':
        return mxnet.cpu()
    elif config['defaults']['mxnet_device'] == 'gpu':
        return mxnet.gpu()



from .factor import Factor


class Module(Factor):
    """
    Modules are a combined Model/Posterior + Inference algorithm object in MXFusion. They act as Factors and are defined as such during model definition, producing Random Variables like a typical Distribution. During inference, instead of the Module having a closed form solution, the outside inference algorithm will call the module's internal inference algorithm to perform modular inference.
    """

    def __init__(self, graph, inputs, outputs, input_names, output_names):
        self._graph = graph

        super(Module, self).__init__(inputs=inputs, outputs=outputs, input_names=input_names, output_names=output_names)

    def compute_log_prob(self, F, targets, conditionals=None, constants=None):
        raise NotImplementedError

    def draw_samples(self, F, num_samples=1, targets=None, conditionals=None,
                     constants=None):
        raise NotImplementedError



from abc import ABC, abstractmethod
from ...common.config import get_default_MXNet_mode


class VariableTransformation(ABC):
    """
    Abstract class for transformations and constraints applied to Variables.
    """
    @abstractmethod
    def transform(self, var, F=None, dtype=None):
        """
        Forward transformation.

        :param var: Variable to be transformed.
        :type var: mx.ndarray or mx.sym
        :param F: Mode to run MxNet in.
        :type F: mxnet.ndarray or mxnet.symbol
        :param dtype: data type.
        :type dtype: e.g. np.float32
        """
        pass

    @abstractmethod
    def inverseTransform(self, out_var, F=None, dtype=None):
        """
        Inverse transformation.

        :param out_var: Variable to be transformed.
        :type out_var: mx.ndarray or mx.sym
        :param F: Mode to run MxNet in.
        :type F: mxnet.ndarray or mxnet.symbol
        :param dtype: data type.
        :type dtype: e.g. np.float32
        """
        pass

class Softplus(VariableTransformation):

    """
    Transformation to apply the Softplus and inverse Softplus functions.
    f = log(1+exp(x))+c
    f^-1 = log(exp(x-c)-1)
    """
    def __init__(self, offset):
        self._offset = offset

    def transform(self, var, F=None, dtype=None):
        """
        Forward transformation.

        :param var: Variable to be transformed.
        :type var: mx.ndarray or mx.sym
        :param F: Mode to run MxNet in.
        :type F: mxnet.ndarray or mxnet.symbol
        :param dtype: data type.
        :type dtype: e.g. np.float32
        """
        F = get_default_MXNet_mode() if F is None else F
        return F.Activation(var, act_type='softrelu') + self._offset

    def inverseTransform(self, out_var, F=None, dtype=None):
        """
        Inverse transformation.

        :param out_var: Variable to be transformed.
        :type out_var: mx.ndarray or mx.sym
        :param F: Mode to run MxNet in.
        :type F: mxnet.ndarray or mxnet.symbol
        :param dtype: data type.
        :type dtype: e.g. np.float32
        """
        F = get_default_MXNet_mode() if F is None else F
        # TODO: make a customized operator for this computation to be overflow
        # robust.
        return F.log(F.expm1(out_var - self._offset))


class PositiveTransformation(Softplus):
    """
    Transformation positively constrain a Variable. Wrapper for the Softplus transformation with offset 0.
    """
    def __init__(self):
        """
        Initializes as Softplus transformation with 0 offset.
        """
        super(PositiveTransformation, self).__init__(offset=0.)



from abc import ABC, abstractmethod
from mxnet.gluon import HybridBlock
from ..components.variables import VariableType
from ..components.variables import add_sample_dimension_arrays
from ..util.inference import variables_to_UUID


class ObjectiveBlock(HybridBlock):
    """
    Wrapping the MXNet codes with a MXNet Gluon HybridBlock for execution.

    :param infr_method: the pointer to a function that contains the actual MXNet codes.
    :type infr_method: a pointer to a function
    :param constants: the variables with constant values
    :type constants: {Variable UUID: int or float or mxnet.ndarray}
    :param data_def: a list of variable UUID, which corresponds to the order of variables expected as the positional arguments in "hybrid_forward".
    :type data_def: [UUID]
    :param var_trans: the transformations applied variables
    :type var_trains: {UUID: VariableTransformation}
    :param var_ties: A dictionary of variables that are tied and use the MXNet Parameter of the dict value uuid.
    :type var_ties: { UUID of source variable: UUID of target variable}
    :param excluded: a set of variables excluded from being set as Block parameters.
    :type excluded: set(UUID) or [UUID]
    :param prefix: the prefix of this Gluon block
    :type prefix: str
    :param params: the ParameterDict of this Gluon block
    :type params: mxnet.gluon.ParameterDict
    """
    def __init__(self, infr_method, constants, data_def, var_trans, var_ties,
                 excluded, prefix='', params=None):
        super(ObjectiveBlock, self).__init__(prefix=prefix, params=params)
        self._infr_method = infr_method
        self._constants = constants
        self._data_def = data_def
        self._var_trans = var_trans
        self._var_ties = var_ties
        for name in params:
            if name not in excluded:
                setattr(self, name, params.get(name))

    def hybrid_forward(self, F, x, *args, **kw):

        for to_uuid, from_uuid in self._var_ties.items():
            kw[to_uuid] = kw[from_uuid]
        data = {k: v for k, v in zip(self._data_def, args)}
        data = add_sample_dimension_arrays(F, data)
        for k, v in self._var_trans.items():
            kw[k] = v.transform(kw[k], F=F)
        kw = add_sample_dimension_arrays(F, kw)
        constants = add_sample_dimension_arrays(F, self._constants)
        obj = self._infr_method.compute(F=F, data=data, parameters=kw,
                                        constants=constants)
        return obj


class InferenceAlgorithm(ABC):
    """
    The abstract class for an inference algorithm. A concrete inference
    algorithm will inherit this class and overload the "compute" function with
    the actual computation logic.

    :param model: the definition of the probabilistic model
    :type model: Model
    :param observed: A list of observed variables
    :type observed: [Variable]
    :param extra_graphs: a list of extra FactorGraph used in the inference
                         algorithm.
    :type extra_graphs: [FactorGraph]
    """

    def __init__(self, model, observed, extra_graphs=None):
        self._model_graph = model
        self._extra_graphs = extra_graphs if extra_graphs is not None else []
        self._graphs = [model] if extra_graphs is None else \
            [model] + extra_graphs
        self._observed = observed
        self._observed_uuid = variables_to_UUID(observed)
        self._observed_names = [v.name for v in observed]

    @property
    def observed_variables(self):
        return self._observed

    @property
    def observed_variable_UUIDs(self):
        return self._observed_uuid

    @property
    def observed_variable_names(self):
        return self._observed_names

    @property
    def model(self):
        """
        return the model definition.
        """
        return self._model_graph

    @property
    def graphs(self):
        return self._graphs

    def create_executor(self, data_def, params, var_ties, rv_scaling=None):
        """
        Create a MXNet Gluon block to carry out the computation.

        :param data_def: a list of unique ID of data variables. The order of
            variables in the list corresponds to the order of variable in the
            positional arguments when calling the Gluon Block.
        :type data_def: [UUID of Variable (str)]
        :param params: The parameters of model involved in inference.
        :type params: InferenceParameters
        :param var_ties: A dictionary of variables that are tied and use the MXNet Parameter of the dict value uuid.
        :type var_ties: { UUID of source variable: UUID of target variable}
        :param rv_scaling: the scaling factor of random variables
        :type rv_scaling: {UUID: scaling factor}
        :returns: the Gluon block computing the outcome of inference
        :rtype: mxnet.gluon.HybridBlock
        """
        excluded = set()
        var_trans = {}
        rv_scaling = {} if rv_scaling is None else rv_scaling
        for g in self._graphs:
            for v in g.variables.values():
                if v.type == VariableType.PARAMETER and v.transformation is not None:
                    var_trans[v.uuid] = v.transformation
                if v.type == VariableType.PARAMETER and v.isInherited:
                    excluded.add(v.uuid)
                if v.type == VariableType.RANDVAR:
                    if v.uuid in rv_scaling:
                        v.factor.log_pdf_scaling = rv_scaling[v.uuid]
                    else:
                        v.factor.log_pdf_scaling = 1
        block = ObjectiveBlock(infr_method=self, params=params.param_dict,
                               constants=params.constants,
                               data_def=data_def, var_trans=var_trans,
                               var_ties=var_ties, excluded=excluded)
        return block

    @abstractmethod
    def compute(self, F, data, parameters, constants):
        """
        The abstract method for the computation of the inference algorithm

        :param F: the execution context (mxnet.ndarray or mxnet.symbol)
        :type F: Python module
        :param data: the data variables for inference
        :type data: {Variable: mxnet.ndarray.ndarray.NDArray or
            mxnet.symbol.symbol.Symbol}
        :param parameters: the parameters for inference
        :type parameters: {Variable: mxnet.ndarray.ndarray.NDArray or
            mxnet.symbol.symbol.Symbol}
        :param constants: the constants for inference
        :type parameters: {Variable: mxnet.ndarray.ndarray.NDArray or
            mxnet.symbol.symbol.Symbol}
        :returns: the outcome of the inference algorithm
        :rtype: mxnet.ndarray.ndarray.NDArray or mxnet.symbol.symbol.Symbol
        """
        raise NotImplementedError



import mxnet as mx
from .grad_loop import GradLoop
from mxnet.gluon.data import ArrayDataset


class MinibatchInferenceLoop(GradLoop):
    """
    The class for the main loop for minibatch gradient-based optimization. The
    *batch_size* specifies the size of mini-batch is used in mini-batch
    inference. *rv_scaling* is used to re-scale the log-likelihood over data
    due to the sub-sampling of the training set. The scaling should be applied
    to all the observed and hidden random variables that are impacted by data
    sub-sampling. The scaling factor is the size of training set dividing the
    batch size.

    :param batch_size: the size of minibatch for optimization
    :type batch_size: int
    :param rv_scaling: the scaling factor of random variables
    :type rv_scaling: {Variable: scaling factor}
    """
    def __init__(self, batch_size=100, rv_scaling=None):
        super(MinibatchInferenceLoop, self).__init__()
        self.batch_size = batch_size
        self.rv_scaling = {v.uuid: s for v, s in rv_scaling.items()} \
            if rv_scaling is not None else rv_scaling

    def run(self, infr_executor, data, param_dict, ctx, optimizer='adam',
            learning_rate=1e-3, max_iter=2000, verbose=False):
        """
        :param infr_executor: The MXNet function that computes the training objective.
        :type infr_executor: MXNet Gluon Block
        :param data: a list of observed variables
        :type data: [mxnet.ndarray]
        :param param_dict: The MXNet ParameterDict for Gradient-based optimization
        :type param_dict: mxnet.gluon.ParameterDict
        :param ctx: MXNet context
        :type ctx: mxnet.cpu or mxnet.gpu
        :param optimizer: the choice of optimizer (default: 'adam')
        :type optimizer: str
        :param learning_rate: the learning rate of the gradient optimizer (default: 0.001)
        :type learning_rate: float
        :param max_iter: the maximum number of iterations of gradient optimization
        :type max_iter: int
        :param verbose: whether to print per-iteration messages.
        :type verbose: boolean
        """

        if isinstance(data, mx.gluon.data.DataLoader):
            data_loader = data
        else:
            data_loader = mx.gluon.data.DataLoader(
                ArrayDataset(*data), batch_size=self.batch_size, shuffle=True,
                last_batch='rollover')
        trainer = mx.gluon.Trainer(param_dict,
                                   optimizer=optimizer,
                                   optimizer_params={'learning_rate':
                                                     learning_rate})
        for e in range(max_iter):
            L_e = 0
            n_batches = 0
            for i, data_batch in enumerate(data_loader):
                with mx.autograd.record():
                    loss = infr_executor(mx.nd.zeros(1, ctx=ctx), *data_batch)
                    loss.backward()
                if verbose:
                    print('\repoch {} Iteration {} logL: {}\t\t'.format(
                          e + 1, i + 1, loss.asscalar() / self.batch_size),
                          end='')
                trainer.step(batch_size=self.batch_size,
                             ignore_stale_grad=True)
                L_e += loss.asscalar() / self.batch_size
                n_batches += 1
            if verbose:
                print('epoch-loss: {} '.format(L_e / n_batches))



from .inference import Inference
from .batch_loop import BatchInferenceLoop


class GradBasedInference(Inference):
    """
    The abstract class for gradient-based inference methods.
    An inference method consists of a few components: the applied inference algorithm, the model definition (optionally a definition of posterior
    approximation), the inference parameters.

    :param inference_algorithm: The applied inference algorithm
    :type inference_algorithm: InferenceAlgorithm
    :param graphs: a list of graph definitions required by the inference method. It includes the model definition and necessary posterior approximation.
    :type graphs: [FactorGraph]
    :param observed: A list of observed variables
    :type observed: [Variable]
    :param grad_loop: The reference to the main loop of gradient optmization
    :type grad_loop: GradLoop
    :param constants: Specify a list of model variables as constants
    :type constants: {Variable: mxnet.ndarray}
    :param hybridize: Whether to hybridize the MXNet Gluon block of the inference method.
    :type hybridize: boolean
    :param dtype: data type for internal numberical representation
    :type dtype: {numpy.float64, numpy.float32, 'float64', 'float32'}
    :param context: The MXNet context
    :type context: {mxnet.cpu or mxnet.gpu}
    """
    def __init__(self, inference_algorithm, grad_loop=None, constants=None,
                 hybridize=False, dtype=None, context=None):
        if grad_loop is None:
            grad_loop = BatchInferenceLoop()
        super(GradBasedInference, self).__init__(
            inference_algorithm=inference_algorithm, constants=constants,
            hybridize=hybridize, dtype=dtype, context=context)
        self._grad_loop = grad_loop

    def create_executor(self):
        """
        Return a MXNet Gluon block responsible for the execution of the inference method.
        """
        from .minibatch_loop import MinibatchInferenceLoop
        if isinstance(self._grad_loop, MinibatchInferenceLoop):
            rv_scaling = self._grad_loop.rv_scaling
        else:
            rv_scaling = None
        infr = self._infr_alg.create_executor(
            data_def=self.observed_variable_UUIDs, params=self.params,
            var_ties=self.params.var_ties, rv_scaling=rv_scaling)
        if self._hybridize:
            infr.hybridize()
        infr.initialize(ctx=self.mxnet_context)
        return infr

    def run(self, optimizer='adam', learning_rate=1e-3, max_iter=2000,
            verbose=False, **kw):
        """
        Run the inference method.

        :param optimizer: the choice of optimizer (default: 'adam')
        :type optimizer: str
        :param learning_rate: the learning rate of the gradient optimizer (default: 0.001)
        :type learning_rate: float
        :param max_iter: the maximum number of iterations of gradient optimization
        :type max_iter: int
        :param verbose: whether to print per-iteration messages.
        :type verbose: boolean
        :param \**kwargs: The keyword arguments specify the data for inferences. The key of each argument is the name of the corresponding
            variable in model definition and the value of the argument is the data in numpy array format.
        """
        data = [kw[v] for v in self.observed_variable_names]
        self.initialize(**kw)

        infr = self.create_executor()
        return self._grad_loop.run(
            infr_executor=infr, data=data, param_dict=self.params.param_dict,
            ctx=self.mxnet_context, optimizer=optimizer,
            learning_rate=learning_rate, max_iter=max_iter, verbose=verbose)



from ..variables import Variable
from .univariate import UnivariateDistribution
from .distribution import LogPDFDecorator, DrawSamplesDecorator
from ...util.customop import broadcast_to_w_samples
from ..variables import get_num_samples, is_sampled_array
from ...common.config import get_default_MXNet_mode
from ...common.exceptions import InferenceError


class CategoricalLogPDFDecorator(LogPDFDecorator):

    def _wrap_log_pdf_with_broadcast(self, func):
        def log_pdf_broadcast(self, F, **kw):
            """
            Computes the logrithm of the probability density/mass function (PDF/PMF) of the distribution.

            :param F: the MXNet computation mode (mxnet.symbol or mxnet.ndarray)
            :param kw: the dict of input and output variables of the distribution
            :type kw: {name: MXNet NDArray or MXNet Symbol}
            :returns: log pdf of the distribution
            :rtypes: MXNet NDArray or MXNet Symbol
            """
            variables = {name: kw[name] for name, _ in self.inputs}
            variables['random_variable'] = kw['random_variable']
            rv_shape = variables['random_variable'].shape[1:]

            nSamples = max([get_num_samples(F, v) for v in variables.values()])
            full_shape = list(rv_shape)
            full_shape[self.axis] = self.num_classes
            full_shape = tuple(full_shape)
            full_shape = (nSamples,) + full_shape

            variables = {
                name: broadcast_to_w_samples(F, v, full_shape[:-1]+(v.shape[-1],)) for name, v in
                variables.items()}
            res = func(self, F=F, **variables)
            return res
        return log_pdf_broadcast


class CategoricalDrawSamplesDecorator(DrawSamplesDecorator):

    def _wrap_draw_samples_with_broadcast(self, func):
        def draw_samples_broadcast(self, F, rv_shape, num_samples=1,
                                   always_return_tuple=False, **kw):
            """
            Draw a number of samples from the distribution.

            :param F: the MXNet computation mode (mxnet.symbol or mxnet.ndarray)
            :param rv_shape: the shape of each sample
            :type rv_shape: tuple
            :param nSamples: the number of drawn samples (default: one)
            :int nSamples: int
            :param always_return_tuple: Whether return a tuple even if there is only one variables in outputs.
            :type always_return_tuple: boolean
            :param kw: the dict of input variables of the distribution
            :type kw: {name: MXNet NDArray or MXNet Symbol}
            :returns: a set samples of the distribution
            :rtypes: MXNet NDArray or MXNet Symbol or [MXNet NDArray or MXNet Symbol]
            """
            rv_shape = list(rv_shape.values())[0]
            variables = {name: kw[name] for name, _ in self.inputs}

            isSamples = any([is_sampled_array(F, v) for v in variables.values()])
            if isSamples:
                num_samples_inferred = max([get_num_samples(F, v) for v in
                                         variables.values()])
                if num_samples_inferred != num_samples:
                    raise InferenceError("The number of samples in the nSamples argument of draw_samples of Gaussian process has to be the same as the number of samples given to the inputs. nSamples: "+str(num_samples)+" the inferred number of samples from inputs: "+str(num_samples_inferred)+".")
            full_shape = list(rv_shape)
            full_shape[self.axis] = self.num_classes
            full_shape = tuple(full_shape)
            full_shape = (num_samples,) + full_shape

            variables = {
                name: broadcast_to_w_samples(F, v, full_shape[:-1]+(v.shape[-1],)) for name, v in
                variables.items()}
            res = func(self, F=F, rv_shape=rv_shape, num_samples=num_samples,
                       **variables)
            if always_return_tuple:
                res = (res,)
            return res
        return draw_samples_broadcast


class Categorical(UnivariateDistribution):
    """
    The Categorical distribution.

    :param log_prob: the logarithm of the probability being in each of the classes.
    :type log_prob: Variable
    :param num_classes: the number of classes.
    :type num_classes: int
    :param one_hot_encoding: If true, the random variable is one-hot encoded.
    :type one_hot_encoding: boolean
    :param normalization: If true, a softmax normalization is applied.
    :type normalization: boolean
    :param axis: the axis in which the categorical distribution is assumed (default: -1).
    :type axis: int
    :param rand_gen: the random generator (default: MXNetRandomGenerator).
    :type rand_gen: RandomGenerator
    :param dtype: the data type for float point numbers.
    :type dtype: numpy.float32 or numpy.float64
    :param ctx: the mxnet context (default: None/current context).
    :type ctx: None or mxnet.cpu or mxnet.gpu
    """
    def __init__(self, log_prob, num_classes, one_hot_encoding=False,
                 normalization=True, axis=-1, rand_gen=None, dtype=None,
                 ctx=None):
        if not isinstance(log_prob, Variable):
            log_prob = Variable(value=log_prob)
        inputs = [('log_prob', log_prob)]
        input_names = ['log_prob']
        output_names = ['random_variable']
        super(Categorical, self).__init__(
            inputs=inputs, outputs=None,
            input_names=input_names,
            output_names=output_names,
            rand_gen=rand_gen, dtype=dtype,
            ctx=ctx)
        if axis != -1:
            raise NotImplementedError("The Categorical distribution currently only supports the last dimension to be the class label dimension, i.e., axis == -1.")
        self.axis = axis
        self.normalization = normalization
        self.one_hot_encoding = one_hot_encoding
        self.num_classes = num_classes

    def replicate_self(self, attribute_map=None):
        """
        This functions as a copy constructor for the object.
        In order to do a copy constructor we first call ``__new__`` on the class which creates a blank object.
        We then initialize that object using the methods standard init procedures, and do any extra copying of attributes.

        Replicates this Factor, using new inputs, outputs, and a new uuid.
        Used during model replication to functionally replicate a factor into a new graph.

        :param inputs: new input variables of the factor.
        :type inputs: List of tuples of name to node e.g. [('random_variable': Variable y)] or None
        :param outputs: new output variables of the factor.
        :type outputs: List of tuples of name to node e.g. [('random_variable': Variable y)] or None
        """
        replicant = super(Categorical, self).replicate_self(attribute_map=attribute_map)
        replicant.axis = self.axis
        replicant.normalization = self.normalization
        replicant.one_hot_encoding = self.one_hot_encoding
        replicant.num_classes = self.num_classes
        return replicant

    @CategoricalLogPDFDecorator()
    def log_pdf(self, log_prob, random_variable, F=None):
        """
        Computes the logarithm of probabilistic mass function of the Categorical distribution.

        :param F: MXNet computation type <mx.sym, mx.nd>.
        :param log_prob: the logarithm of the probability being in each of the classes.
        :type log_prob: MXNet NDArray or MXNet Symbol
        :param random_variable: the point to compute the logpdf for.
        :type random_variable: MXNet NDArray or MXNet Symbol
        :returns: log pdf of the distribution.
        :rtypes: MXNet NDArray or MXNet Symbol
        """
        F = get_default_MXNet_mode() if F is None else F

        if self.normalization:
            log_prob = F.log_softmax(log_prob, axis=self.axis)

        if self.one_hot_encoding:
            logL = F.sum(F.broadcast_mul(random_variable, log_prob),
                         axis=self.axis) * self.log_pdf_scaling
        else:
            logL = F.pick(log_prob, index=random_variable, axis=self.axis)
            logL = logL * self.log_pdf_scaling
        return logL

    @CategoricalDrawSamplesDecorator()
    def draw_samples(self, log_prob, rv_shape, num_samples=1, F=None):
        """
        Draw a number of samples from the Categorical distribution.

        :param log_prob: the logarithm of the probability being in each of the classes.
        :type log_prob: MXNet NDArray or MXNet Symbol
        :param rv_shape: the shape of each sample.
        :type rv_shape: tuple
        :param num_samples: the number of drawn samples (default: one).
        :int num_samples: int
        :param F: the MXNet computation mode (mxnet.symbol or mxnet.ndarray).
        :returns: a set samples of the Categorical distribution
        :rtypes: MXNet NDArray or MXNet Symbol
        """
        F = get_default_MXNet_mode() if F is None else F

        if self.normalization:
            log_prob = F.log_softmax(log_prob, axis=self.axis)

        log_prob = F.reshape(log_prob, shape=(1, -1, self.num_classes))
        samples = self._rand_gen.sample_multinomial(log_prob, get_prob=False)
        if self.one_hot_encoding:
            samples = F.one_hot(samples, depth=self.num_classes)
        samples = F.reshape(samples, shape=(num_samples,) + rv_shape)
        return samples

    @staticmethod
    def define_variable(log_prob, num_classes, shape=None, one_hot_encoding=False, normalization=True, axis=-1,
                        rand_gen=None, dtype=None, ctx=None):
        """
        Creates and returns a random variable drawn from a Categorical distribution.

        :param log_prob: the logarithm of the probability being in each of the classes.
        :type log_prob: Variable
        :param num_classes: the number of classes.
        :type num_classes: int
        :param shape: the shape of the Categorical variable.
        :type shape: tuple of int
        :param one_hot_encoding: If true, the random variable is one-hot encoded.
        :type one_hot_encoding: boolean
        :param normalization: If true, a softmax normalization is applied.
        :type normalization: boolean
        :param axis: the axis in which the categorical distribution is assumed (default: -1).
        :type axis: int
        :param rand_gen: the random generator (default: MXNetRandomGenerator).
        :type rand_gen: RandomGenerator
        :param dtype: the data type for float point numbers.
        :type dtype: numpy.float32 or numpy.float64
        :param ctx: the mxnet context (default: None/current context).
        :type ctx: None or mxnet.cpu or mxnet.gpu
        :returns: RandomVariable drawn from the Categorical distribution.
        :rtypes: Variable
        """
        cat = Categorical(
            log_prob=log_prob, num_classes=num_classes,
            one_hot_encoding=one_hot_encoding, normalization=normalization,
            axis=axis, rand_gen=rand_gen, dtype=dtype, ctx=ctx)
        cat._generate_outputs(shape=shape)
        return cat.random_variable



from .inference_alg import InferenceAlgorithm
from ..components.variables import Variable, VariableType
from ..models.posterior import Posterior
from ..components.distributions import PointMass
from ..util.inference import variables_to_UUID


class MAP(InferenceAlgorithm):
    """
    The class of the Maximum A Posteriori (MAP) algorithm.

    :param model: the definition of the probabilistic model
    :type model: Model
    :param observed: A list of observed variables
    :type observed: [Variable]
    """
    def __init__(self, model, observed):
        posterior = MAP.create_posterior(model, variables_to_UUID(observed))
        super(MAP, self).__init__(model=model, observed=observed,
                                  extra_graphs=[posterior])

    @property
    def posterior(self):
        """
        return the variational posterior.
        """
        return self._extra_graphs[0]

    @staticmethod
    def create_posterior(model, observed):
        """
        Create the posterior of Maximum A Posteriori (MAP).

        :param model: the model definition
        :type model: Model
        :param observed: the list of observed variables
        :type observed: [UUID]
        :returns: the resulting posterior representation
        :rtype: Posterior
        """
        q = Posterior(model)
        for v in model.get_latent_variables(observed):
            q[v].assign_factor(PointMass(location=Variable(shape=v.shape)))
        return q

    def compute(self, F, data, parameters, constants):
        """
        The method for the computation of the inference algorithm

        :param F: the execution context (mxnet.ndarray or mxnet.symbol)
        :type F: Python module
        :param data: the data variables for inference
        :type data: {Variable: mxnet.ndarray.ndarray.NDArray or
            mxnet.symbol.symbol.Symbol}
        :param parameters: the parameters for inference
        :type parameters: {Variable: mxnet.ndarray.ndarray.NDArray or
            mxnet.symbol.symbol.Symbol}
        :param constants: the constants for inference
        :type parameters: {Variable: mxnet.ndarray.ndarray.NDArray or
            mxnet.symbol.symbol.Symbol}
        :returns: the outcome of the inference algorithm
        :rtype: mxnet.ndarray.ndarray.NDArray or mxnet.symbol.symbol.Symbol
        """
        knowns = data.copy()
        knowns.update(constants)

        for v in self.model.variables.values():
            if v.type == VariableType.RANDVAR and v not in data:
                # TODO: check self._post_graph[v].factor is a PointMass distribution
                knowns[v.uuid] = parameters[self.posterior[v].factor.location.uuid]
            elif v.type == VariableType.PARAMETER and v not in data and v not in constants:
                if not v.isInherited:
                    knowns[v.uuid] = parameters[v.uuid]

        logL = self.model.compute_log_prob(F=F, targets=knowns,
                                           constants=constants)
        return -logL



from .factor_graph import FactorGraph
from ..components import VariableType


class Model(FactorGraph):
    """
    The model defines a factor graph over a set of variables for use in inference.
    """

    def __init__(self, name=None, verbose=False):
        """
        Model object constructor.

        :param name: optional parameter to name the model for easier reference.
        """
        super(Model, self).__init__(name=name, verbose=verbose)

    def get_latent_variables(self, observed):
        """
        Get the latent variables of the model.

        :param observed: a list of observed variables.
        :type observed: [UUID]
        :returns: the list of latent variables.
        :rtype: [Variable]
        """
        return [v for v in self.variables.values() if v.type == VariableType.RANDVAR and v.uuid not in observed]

    def _replicate_class(self, **kwargs):
        """
        Returns a new instance of the derived FactorGraph's class.
        """
        return Model(**kwargs)



from uuid import uuid4
from ..common.exceptions import ModelSpecificationError


class ModelComponent(object):
    """
    The building block of a Model in MXFusion.

    ModelComponents exist in one of two modes.

    **Mode 1 - Bi-directional mode**

    If a node is not attached to a FactorGraph, it maintains a list of all of its predecessors and successors directly.
    These are stored in the ``self._predecessors`` and ``self._successors`` methods.

    **Mode 2 - Graph mode**

    If a node is attached to a FactorGraph, it does not store direct references to its successors and predecessors.
    When accessed, the predecessors/successors properties directly query the graph they are attached to to find out what the respective neighbor nodes are.
    """

    def __init__(self):
        self.name = None
        self._uuid = str(uuid4()).replace('-', '_')
        self._parent_graph = None
        self._successors = []  # either [('name', Variable), ('factor', Factor), ...]
        self._predecessors = []
        self.attributes = []

    @property
    def uuid(self):
        """
        Return the UUID of this graph
        """
        return self._uuid

    def __hash__(self):
        return self._uuid.__hash__()

    def __eq__(self, other):
        return self._uuid.__hash__() == other.__hash__()

    def __repr__(self):
        return self.uuid

    @property
    def graph(self):
        """
        Return the object's graph
        """
        return self._parent_graph

    @graph.setter
    def graph(self, graph):
        """
        Attaches the node to a graph, switching from Bidirectional mode to Graph mode if it is not already in Graph mode.

        A node cannot be re-attached to a different graph once it is attached. Use the ``replicate()`` functionality if you need to do this.

        :param graph: The ``components_graph`` of the ``FactorGraph`` this node is attaching to.
        :type graph: networkx.DiGraph
        """
        if self._parent_graph is not None:
            if self._parent_graph == graph:
                return
            elif graph is not None:
                raise ModelSpecificationError("Trying to reset a variables graph is bad!")
        self._parent_graph = graph
        if self._parent_graph is not None:
            self._parent_graph.add_node(self)

        self.predecessors = self._predecessors
        self.successors = self._successors
        self._update_attributes()

        self._predecessors = []
        self._successors = []

    def _update_attributes(self):
        """
        Adds all of a node's attributes to its graph.
        """
        for a in self.attributes:
            self.graph.add_node(a)

    def _align_graph_modes(self, edge_nodes):
        """
        This function will update the current node and all nodes passed in to be in Graph mode if any of edge_nodes are in Graph mode.

        :param edge_nodes: All the nodes to align to the same graph mode. I.E. predecessors or successors.
        :type edge_nodes: List of tuples of name to node e.g. [('random_variable': Variable y)]
        """

        if self.graph is None and any([node.graph is not None for _, node in edge_nodes]):
            # Put self into the graph, put all the other nodes in that graph (if more than one just error).
            graphs = set([node.graph for _, node in edge_nodes if node.graph is not None])
            if len(graphs) > 1:
                raise ModelSpecificationError("Too many graphs!")
            graph = list(graphs)[0]
            self.graph = graph
            for _, node in edge_nodes:
                node.graph = graph

    @property
    def successors(self):
        """
        Return a list of nodes pointed to by the edges of this node.
        """
        if self.graph is not None:
            succ = [(e['name'], v) for v, e in self.graph.succ[self].items()]
            return succ
        else:
            return self._successors

    @successors.setter
    def successors(self, successors):
        """
        Sets this node's successors to those passed in.

        :param successors: List of tuples of name to node e.g. [('random_variable': Variable y)].
        :type successors: List of tuples of name to node e.g. [('random_variable': Variable y)]
        """
        def add_predecessor(successor, predecessor, successor_name):
            if successor.graph is None:
                successor._predecessors.append((successor_name, predecessor))
            if successor.graph is not None:
                raise ModelSpecificationError("Internal Error. Cannot add predecessor when a component is attached to a graph.")

        self._align_graph_modes(successors)
        if self.graph is not None:
            for _, successor in self.successors:
                self.graph.remove_edge(self, successor)
            for name, successor in successors:
                successor.graph = self.graph
                self.graph.add_edge(self, successor, name=name)
        else:
            self._successors = successors
            for name, successor in successors:
                add_predecessor(successor, self, name)


    @property
    def predecessors(self):
        """
        Return a list of nodes whose edges point into this node.

        :param predecessors: List of tuples of name to node e.g. [('random_variable': Variable y)]
        :type predecessors: List of tuples of name to node e.g. [('random_variable': Variable y)]
        """
        if self.graph is not None:
            pred = [(e['name'], v) for v, e in self.graph.pred[self].items()]
            return pred
        else:
            return self._predecessors

    @predecessors.setter
    def predecessors(self, predecessors):
        """
        Sets this node's predecessors to be those passed in.

        :param predecessors: List of tuples of name to node e.g. [('random_variable': Variable y)]
        :type predecessors: List of tuples of name to node e.g. [('random_variable': Variable y)]
        """
        def add_successor(predecessor, successor, predecessor_name):
            if predecessor.graph is None:
                predecessor._successors.append((predecessor_name, successor))
            if predecessor.graph is not None:
                raise ModelSpecificationError("Internal Error. Cannot add a successor when a component is attached to a graph.")

        self._align_graph_modes(predecessors)
        if self.graph is not None:
            for _, predecessor in self.predecessors:
                self.graph.remove_edge(predecessor, self)
            for name, predecessor in predecessors:
                predecessor.graph = self.graph
                self.graph.add_edge(predecessor, self, name=name)
        else:
            self._predecessors = predecessors
            for name, predecessor in predecessors:
                add_successor(predecessor, self, name)

    def _replicate_self_with_attributes(self, var_map):
        """
        Replicates self if not in ``var_map``. Also replicates all of self's attributes.

        :param var_map: A mapping from the original model's components to the replicated components.
        :type var_map: {original_node: new_node}
        :rtype: ModelComponent
        """
        var_map = var_map if var_map is not None else {}
        if self in var_map:
            return var_map[self]

        attributes_map = {}
        for a in self.attributes:
            if a in var_map:
                attributes_map[a] = var_map[a]
            else:
                attributes_map[a] = a.replicate_self()
                var_map[a] = attributes_map[a]
        replicated_component = self.replicate_self(attributes_map)
        var_map[self] = replicated_component
        return replicated_component

    def _replicate_neighbors(self, var_map, neighbors, recurse_type, replication_function):
        """
        Helper function that returns a replicated list of neighbors based on the recurse_type passed in.

        :param var_map: A mapping from the original model's components to the replicated components.
        :type var_map: {original_node: new_node}
        :param neighbors: Dictionary containing the list of a node's neighbors in one direction (predecessors or successors).
        :type neighbors: List of tuples of name to node e.g. [('random_variable': Variable y)]
        :param recurse_type: Parameter that decides how to replicate the neighbor nodes. Must be one of: 'recursive', 'one_level', or None.
        :type recurse_type: String or None
        :param replication_function: A function that takes in a ModelComponent and returns an answer for how to replicate that node's predecessors and successors.
        :type replication_function: function

        """
        if recurse_type == 'recursive':
            replicated_neighbors = [(name, i.replicate(var_map=var_map, replication_function=replication_function))
                                   for name, i in neighbors]
        elif recurse_type == 'one_level':
            replicated_neighbors = [(name, i._replicate_self_with_attributes(var_map=var_map))
                                   for name, i in neighbors]
        elif recurse_type is None:
            replicated_neighbors = []
        else:
            raise ModelSpecificationError("Parameter 'recurse_type' must be 'recursive', 'one_level', or None.")
        return replicated_neighbors

    def replicate(self, var_map=None, replication_function=None):
        """
        Replicates this component and its neighbors based on the replication_function logic passed in.

        :param var_map: A mapping from the original model's components to the replicated components. This is used to track which components
            have already been replicated in a dynamic programming style.
        :type var_map: {original_node: new_node}
        :param replication_function: A function that takes in a ModelComponent and returns an answer for how to replicate that node's predecessors and successors.
        :type replication_function: function
        """
        var_map = var_map if var_map is not None else {}
        if self in var_map:
            return var_map[self]
        replicated_component = self._replicate_self_with_attributes(var_map)

        if replication_function is not None:
            pred_recursion, succ_recursion = replication_function(self)
        else:
            pred_recursion, succ_recursion = None, None
        predecessors = self._replicate_neighbors(var_map, self.predecessors, pred_recursion, replication_function)
        successors = self._replicate_neighbors(var_map, self.successors, succ_recursion, replication_function)
        replicated_component.predecessors = predecessors
        replicated_component.successors = successors

        return replicated_component



"""This module contains implementations of Gaussian Processes for MXFusion.

Submodules
==========

.. autosummary::
    :toctree: _autosummary

    add_kernel
    kernel
    linear
    rbf
    static
    stationary
"""

__all__ = ['add_kernel', 'kernel', 'linear', 'rbf', 'static', 'stationary']

from .add_kernel import AddKernel
from .rbf import RBF
from .linear import Linear
from .static import Bias, White



from ..common.exceptions import InferenceError
from ..components.variables import Variable
from .variational import StochasticVariationalInference
from .inference_alg import InferenceAlgorithm
from .inference import TransferInference
from .map import MAP


class ForwardSamplingAlgorithm(InferenceAlgorithm):
    """
    The class of the forward sampling algorithm.

    :param num_samples: the number of samples used in estimating the variational lower bound
    :type num_samples: int
    :param model_graph: the definition of the probabilistic model
    :type model_graph: Model
    :param target_variables: (optional) the target variables to sample
    :type target_variables: [UUID]
    :param observed: A list of observed variables
    :type observed: [Variable]
    """

    def __init__(self, num_samples, model, observed, target_variables=None):
        super(ForwardSamplingAlgorithm, self).__init__(model=model,
                                                       observed=observed)
        self.num_samples = num_samples
        self._target_variables = target_variables

    def compute(self, F, data, parameters, constants):
        """
        The method for the computation of the inference algorithm

        :param F: the execution context (mxnet.ndarray or mxnet.symbol)
        :type F: Python module
        :param data: the data variables for inference
        :type data: {Variable: mxnet.ndarray.ndarray.NDArray or
            mxnet.symbol.symbol.Symbol}
        :param parameters: the parameters for inference
        :type parameters: {Variable: mxnet.ndarray.ndarray.NDArray or
            mxnet.symbol.symbol.Symbol}
        :param constants: the constants for inference
        :type parameters: {Variable: mxnet.ndarray.ndarray.NDArray or
            mxnet.symbol.symbol.Symbol}
        :returns: the outcome of the inference algorithm
        :rtype: mxnet.ndarray.ndarray.NDArray or mxnet.symbol.symbol.Symbol
        """
        knowns = data.copy()
        knowns.update(parameters)
        knowns.update(constants)
        samples = self.model.draw_samples(
            F=F, targets=self._target_variables, conditionals=knowns,
            num_samples=self.num_samples, constants=constants)
        if self._target_variables is not None:
            samples = {v: samples[v] for v in self._target_variables}
        samples = {k: v for k, v in samples.items()}
        return samples


class ForwardSampling(TransferInference):
    """
    Inference method of forward sampling.

    :param num_samples: the number of samples used in estimating the variational lower bound
    :type num_samples: int
    :param model_graph: the definition of the probabilistic model
    :type model_graph: Model
    :param observed: A list of observed variables
    :type observed: [Variable]
    :param var_ties: A dictionary of variables that are tied together and use the MXNet Parameter of the dict value's uuid.
    :type var_ties: { UUID to tie from : UUID to tie to }
    :param infr_params: list or single of InferenceParameters objects from previous Inference runs.
    :type infr_params: InferenceParameters or [InferenceParameters]
    :param target_variables: (optional) the target variables to sample
    :type target_variables: [Variable]
    :param constants: Specify a list of model variables as constants
    :type constants: {Variable: mxnet.ndarray}
    :param hybridize: Whether to hybridize the MXNet Gluon block of the inference method.
    :type hybridize: boolean
    :param dtype: data type for internal numberical representation
    :type dtype: {numpy.float64, numpy.float32, 'float64', 'float32'}
    :param context: The MXNet context
    :type context: {mxnet.cpu or mxnet.gpu}
    """
    def __init__(self, num_samples, model, observed, var_tie, infr_params,
                 target_variables=None, hybridize=False, constants=None,
                 dtype=None, context=None):
        if target_variables is not None:
            target_variables = [v.uuid for v in target_variables if
                                isinstance(v, Variable)]

        infr = ForwardSamplingAlgorithm(
            num_samples=num_samples, model=model, observed=observed,
            target_variables=target_variables)
        super(ForwardSampling, self).__init__(
            inference_algorithm=infr, var_tie=var_tie, infr_params=infr_params,
            constants=constants, hybridize=hybridize, dtype=dtype,
            context=context)


def merge_posterior_into_model(model, posterior, observed):
    new_model, var_map = model.clone()
    for lv in model.get_latent_variables(observed):
        v = posterior.extract_distribution_of(posterior[lv])
        new_model.replace_subgraph(new_model[v], v)
    return new_model, var_map


class VariationalPosteriorForwardSampling(ForwardSampling):
    """
    The forward sampling method for variational inference.

    :param num_samples: the number of samples used in estimating the variational lower bound
    :type num_samples: int
    :param inherited_inference: the inference method of which the model and inference results are taken
    :type inherited_inference: SVIInference or SVIMiniBatchInference
    :param observed: A list of observed variables
    :type observed: [Variable]
    :param target_variables: (optional) the target variables to sample
    :type target_variables: [Variable]
    :param constants: Specify a list of model variables as constants
    :type constants: {Variable: mxnet.ndarray}
    :param hybridize: Whether to hybridize the MXNet Gluon block of the inference method.
    :type hybridize: boolean
    :param dtype: data type for internal numberical representation
    :type dtype: {numpy.float64, numpy.float32, 'float64', 'float32'}
    :param context: The MXNet context
    :type context: {mxnet.cpu or mxnet.gpu}
    """
    def __init__(self, num_samples, observed,
                 inherited_inference, target_variables=None,
                 hybridize=False, constants=None, dtype=None, context=None):
        if not isinstance(inherited_inference.inference_algorithm,
                          (StochasticVariationalInference, MAP)):
            raise InferenceError('inherited_inference needs to be a subclass of SVIInference or SVIMiniBatchInference.')

        m = inherited_inference.inference_algorithm.model
        q = inherited_inference.inference_algorithm.posterior

        model_graph, var_map = merge_posterior_into_model(
            m, q, observed=inherited_inference.observed_variables)

        super(VariationalPosteriorForwardSampling, self).__init__(
            num_samples=num_samples, model=model_graph,
            observed=observed,
            var_tie={}, infr_params=inherited_inference.params,
            target_variables=target_variables, hybridize=hybridize,
            constants=constants, dtype=dtype, context=context)



"""The main module for MXFusion.

Submodules
==========

.. autosummary::
    :toctree: _autosummary

    distributions
    functions
    modules
    variables
    factor
    model_component
"""

__all__ = ['distributions', 'functions', 'modules', 'variables', 'factor', 'model_component']

from .model_component import ModelComponent
from .factor import Factor
from .distributions import Distribution
from .variables import Variable, VariableType
from .functions import MXFusionGluonFunction, FunctionEvaluation
from .modules import Module



"""The main module for MXFusion.

Submodules
==========

.. autosummary::
    :toctree: _autosummary

    components
    models
    inference
    util
"""

__all__ = ['components', 'models', 'inference', 'util']

from .components import distributions, functions, modules, variables
from .models import Model, Posterior
from .components import Variable



import numpy as np
from ....common.config import get_default_MXNet_mode
from ....common.exceptions import InferenceError
from ...variables.variable import Variable
from ....util.customop import broadcast_to_w_samples
from ..distribution import Distribution, LogPDFDecorator, DrawSamplesDecorator
from ...variables.runtime_variable import get_num_samples


class ConditionalGaussianProcessLogPDFDecorator(LogPDFDecorator):

    def _wrap_log_pdf_with_broadcast(self, func):
        def log_pdf_broadcast(self, F, **kw):
            """
            Computes the logrithm of the probability density/mass function (PDF/PMF) of the distribution.

            :param F: the MXNet computation mode (mxnet.symbol or mxnet.ndarray)
            :param kw: the dict of input and output variables of the distribution
            :type kw: {name: MXNet NDArray or MXNet Symbol}
            :returns: log pdf of the distribution
            :rtypes: MXNet NDArray or MXNet Symbol
            """
            variables = {name: kw[name] for name, _ in self.inputs}
            variables['random_variable'] = kw['random_variable']
            rv_shape = variables['random_variable'].shape[1:]

            num_samples = max([get_num_samples(F, v) for v in variables.values()])
            X_shape = (num_samples,) + rv_shape[:-1] + (self.kernel.input_dim,)
            X_cond_shape = (num_samples,) + rv_shape[:-2] + \
                variables['X_cond'].shape[-2:]
            Y_cond_shape = (num_samples,) + rv_shape[:-2] + \
                variables['Y_cond'].shape[-2:]
            rv_shape = (num_samples,) + rv_shape

            variables['X'] = broadcast_to_w_samples(F, variables['X'], X_shape)
            variables['X_cond'] = broadcast_to_w_samples(
                F, variables['X_cond'], X_cond_shape)
            variables['Y_cond'] = broadcast_to_w_samples(
                F, variables['Y_cond'], Y_cond_shape)
            variables['random_variable'] = broadcast_to_w_samples(
                F, variables['random_variable'], rv_shape)
            return func(self, F=F, **variables)
        return log_pdf_broadcast


class ConditionalGaussianProcessDrawSamplesDecorator(DrawSamplesDecorator):

    def _wrap_draw_samples_with_broadcast(self, func):
        def draw_samples_broadcast(self, F, rv_shape, num_samples=1,
                                   always_return_tuple=False, **kw):
            """
            Draw a number of samples from the distribution.

            :param F: the MXNet computation mode (mxnet.symbol or mxnet.ndarray)
            :param rv_shape: the shape of each sample
            :type rv_shape: tuple
            :param nSamples: the number of drawn samples (default: one)
            :int nSamples: int
            :param always_return_tuple: Whether return a tuple even if there is only one variables in outputs.
            :type always_return_tuple: boolean
            :param kw: the dict of input variables of the distribution
            :type kw: {name: MXNet NDArray or MXNet Symbol}
            :returns: a set samples of the distribution
            :rtypes: MXNet NDArray or MXNet Symbol or [MXNet NDArray or MXNet Symbol]
            """
            rv_shape = list(rv_shape.values())[0]
            variables = {name: kw[name] for name, _ in self.inputs}

            num_samples_inferred = max([get_num_samples(F, v) for v in
                                        variables.values()])
            if num_samples_inferred > 1 and num_samples_inferred != num_samples:
                raise InferenceError("The number of samples in the num_samples argument of draw_samples of Gaussian process has to be the same as the number of samples given to the inputs. num_samples: "+str(num_samples)+" the inferred number of samples from inputs: "+str(num_samples_inferred)+".")
            X_shape = (num_samples,) + rv_shape[:-1] + (self.kernel.input_dim,)
            X_cond_shape = (num_samples,) + rv_shape[:-2] + \
                variables['X_cond'].shape[-2:]
            Y_cond_shape = (num_samples,) + rv_shape[:-2] + \
                variables['Y_cond'].shape[-2:]

            variables['X'] = broadcast_to_w_samples(F, variables['X'], X_shape)
            variables['X_cond'] = broadcast_to_w_samples(
                F, variables['X_cond'], X_cond_shape)
            variables['Y_cond'] = broadcast_to_w_samples(
                F, variables['Y_cond'], Y_cond_shape)
            res = func(self, F=F, rv_shape=rv_shape, num_samples=num_samples,
                       **variables)
            if always_return_tuple:
                res = (res,)
            return res
        return draw_samples_broadcast


class ConditionalGaussianProcess(Distribution):
    """
    The Conditional Gaussian process distribution.

    A Gaussian process consists of a kernel function and a mean function (optional). A collection of GP random variables follows a multi-variate
    normal distribution, where the mean is computed from the mean function (zero, if not given) and the covariance matrix is computed from the kernel
    function, both of which are computed given a collection of inputs.

    The conditional Gaussian process is a Gaussian process distribution which is conditioned over a pair of observation X_cond and Y_cond.

    .. math::
       Y \\sim \\mathcal{N}(Y| K_{*c}K_{cc}^{-1}(Y_C - g(X_c)) + g(X), K_{**} - K_{*c}K_{cc}^{-1}K_{*c}^\\top)

    where :math:`g` is the mean function, :math:`K_{**}` is the covariance matrix over :math:`X`, :math:`K_{*c}` is the cross covariance matrix
    between :math:`X` and :math:`X_{c}` and :math:`K_{cc}` is the covariance matrix over :math:`X_{c}`.

    :param X: the input variables on which the random variables are conditioned.
    :type X: Variable
    :param X_cond: the input variables on which the output variables *Y_Cond* are conditioned.
    :type X_cond: Variable
    :param Y_cond: the output variables on which the random variables are conditioned.
    :type Y_cond: Variable
    :param kernel: the kernel of Gaussian process.
    :type kernel: Kernel
    :param mean_func: the mean function of Gaussian process.
    :type mean_func: N/A
    :param rand_gen: the random generator (default: MXNetRandomGenerator).
    :type rand_gen: RandomGenerator
    :param dtype: the data type for float point numbers.
    :type dtype: numpy.float32 or numpy.float64
    :param ctx: the mxnet context (default: None/current context).
    :type ctx: None or mxnet.cpu or mxnet.gpu
    """
    def __init__(self, X, X_cond, Y_cond, kernel, mean_func=None,
                 rand_gen=None, dtype=None, ctx=None):
        if not isinstance(X, Variable):
            X = Variable(value=X)
        inputs = [('X', X), ('X_cond', X_cond), ('Y_cond', Y_cond)] + \
            [(k, v) for k, v in kernel.parameters.items()]
        input_names = [k for k, _ in inputs]
        output_names = ['random_variable']
        super(ConditionalGaussianProcess, self).__init__(
            inputs=inputs, outputs=None, input_names=input_names,
            output_names=output_names, rand_gen=rand_gen, dtype=dtype,
            ctx=ctx)
        self.mean_func = mean_func
        self.kernel = kernel

    @staticmethod
    def define_variable(X, X_cond, Y_cond, kernel, shape=None, mean_func=None,
                        rand_gen=None, minibatch_ratio=1., dtype=None,
                        ctx=None):
        """
        Creates and returns a set of random variable drawn from a Gaussian process.

        :param X: the input variables on which the random variables are conditioned.
        :type X: Variable
        :param X_cond: the input variables on which the output variables *Y_Cond* are conditioned.
        :type X_cond: Variable
        :param Y_cond: the output variables on which the random variables are conditioned.
        :type Y_cond: Variable
        :param kernel: the kernel of Gaussian process.
        :type kernel: Kernel
        :param shape: the shape of the random variable(s) (the default shape is the same shape as *X* but the last dimension is changed to one.)
        :type shape: tuple or [tuple]
        :param mean_func: the mean function of Gaussian process.
        :type mean_func: N/A
        :param rand_gen: the random generator (default: MXNetRandomGenerator).
        :type rand_gen: RandomGenerator
        :param dtype: the data type for float point numbers.
        :type dtype: numpy.float32 or numpy.float64
        :param ctx: the mxnet context (default: None/current context).
        :type ctx: None or mxnet.cpu or mxnet.gpu
        """
        gp = ConditionalGaussianProcess(
            X=X, X_cond=X_cond, Y_cond=Y_cond, kernel=kernel,
            mean_func=mean_func, rand_gen=rand_gen, dtype=dtype, ctx=ctx)
        gp.outputs = [('random_variable',
                      Variable(value=gp, shape=X.shape[:-1] + (1,) if
                               shape is None else shape))]
        return gp.random_variable

    @ConditionalGaussianProcessLogPDFDecorator()
    def log_pdf(self, X, X_cond, Y_cond, random_variable, F=None,
                **kernel_params):
        """
        Computes the logrithm of the probability density function (PDF) of the condtional Gaussian process.

        .. math::
           \\log p(Y| X_c, Y_c, X) = \\log \\mathcal{N}(Y| K_{*c}K_{cc}^{-1}(Y_C - g(X_c)) + g(X), K_{**} - K_{*c}K_{cc}^{-1}K_{*c}^\\top)

        where :math:`g` is the mean function, :math:`K_{**}` is the covariance matrix over :math:`X`, :math:`K_{*c}` is the cross covariance matrix
        between :math:`X` and :math:`X_{c}` and :math:`K_{cc}` is the covariance matrix over :math:`X_{c}`.

        :param X: the input variables on which the random variables are conditioned.
        :type X: MXNet NDArray or MXNet Symbol
        :param X_cond: the input variables on which the output variables *Y_Cond* are conditioned.
        :type X_cond: MXNet NDArray or MXNet Symbol
        :param Y_cond: the output variables on which the random variables are conditioned.
        :type Y_cond: MXNet NDArray or MXNet Symbol
        :param random_variable: the random_variable of which log-PDF is computed.
        :type random_variable: MXNet NDArray or MXNet Symbol
        :param F: the MXNet computation mode (mxnet.symbol or mxnet.ndarray).
        :param \**kernel_params: the set of kernel parameters, provided as keyword arguments.
        :type \**kernel_params: {str: MXNet NDArray or MXNet Symbol}
        :returns: log pdf of the distribution.
        :rtypes: MXNet NDArray or MXNet Symbol
        """
        D = random_variable.shape[-1]
        F = get_default_MXNet_mode() if F is None else F
        K = self.kernel.K(F, X, **kernel_params)
        Kc = self.kernel.K(F, X_cond, X, **kernel_params)
        Kcc = self.kernel.K(F, X_cond, **kernel_params)
        Lcc = F.linalg.potrf(Kcc)
        LccInvKc = F.linalg.trsm(Lcc, Kc)
        cov = K - F.linalg.syrk(LccInvKc, transpose=True)
        L = F.linalg.potrf(cov)
        if self.mean_func is not None:
            random_variable = random_variable - self.mean_func(F, X)
            Y_cond = Y_cond - self.mean_func(F, X_cond)
        LccInvY = F.linalg.trsm(Lcc, Y_cond)
        rv_mean = F.linalg.gemm2(LccInvKc, LccInvY, True, False)
        LinvY = F.sum(F.linalg.trsm(L, random_variable - rv_mean), axis=-1)
        logdet_l = F.linalg.sumlogdiag(F.abs(L))

        return (- logdet_l * D - F.sum(F.square(LinvY) + np.log(2. * np.pi),
                axis=-1) / 2) * self.log_pdf_scaling

    @ConditionalGaussianProcessDrawSamplesDecorator()
    def draw_samples(self, X, X_cond, Y_cond, rv_shape, num_samples=1, F=None,
                     **kernel_params):
        """
        Draw a number of samples from the condtional Gaussian process.

        :param X: the input variables on which the random variables are conditioned.
        :type X: MXNet NDArray or MXNet Symbol
        :param X_cond: the input variables on which the output variables *Y_Cond* are conditioned.
        :type X_cond: MXNet NDArray or MXNet Symbol
        :param Y_cond: the output variables on which the random variables are conditioned.
        :type Y_cond: MXNet NDArray or MXNet Symbol
        :param rv_shape: the shape of each sample.
        :type rv_shape: tuple
        :param num_samples: the number of drawn samples (default: one).
        :int num_samples: int
        :param F: the MXNet computation mode (mxnet.symbol or mxnet.ndarray).
        :param \**kernel_params: the set of kernel parameters, provided as keyword arguments.
        :type \**kernel_params: {str: MXNet NDArray or MXNet Symbol}
        :returns: a set samples of the distribution.
        :rtypes: MXNet NDArray or MXNet Symbol
        """
        F = get_default_MXNet_mode() if F is None else F
        K = self.kernel.K(F, X, **kernel_params)
        Kc = self.kernel.K(F, X_cond, X, **kernel_params)
        Kcc = self.kernel.K(F, X_cond, **kernel_params)
        Lcc = F.linalg.potrf(Kcc)
        LccInvKc = F.linalg.trsm(Lcc, Kc)
        cov = K - F.linalg.syrk(LccInvKc, transpose=True)
        L = F.linalg.potrf(cov)
        if self.mean_func is not None:
            Y_cond = Y_cond - self.mean_func(F, X_cond)
        LccInvY = F.linalg.trsm(Lcc, Y_cond)
        rv_mean = F.linalg.gemm2(LccInvKc, LccInvY, True, False)

        out_shape = (num_samples,) + rv_shape
        L = broadcast_to_w_samples(F, L, out_shape[:-1] + out_shape[-2:-1])

        die = self._rand_gen.sample_normal(
            shape=out_shape, dtype=self.dtype, ctx=self.ctx)
        rv = F.linalg.trmm(L, die) + rv_mean
        if self.mean_func is not None:
            rv = rv + self.mean_func(F, X)
        return rv



from ...common.exceptions import InferenceError
from ..variables import Variable
from .distribution import Distribution, LogPDFDecorator, DrawSamplesDecorator
from ..variables import is_sampled_array, get_num_samples
from ...util.customop import broadcast_to_w_samples


class UnivariateLogPDFDecorator(LogPDFDecorator):

    def _wrap_log_pdf_with_broadcast(self, func):
        def log_pdf_broadcast(self, F, **kws):
            """
            Computes the logrithm of the probability density/mass function
            (PDF/PMF) of the distribution.

            :param F: the MXNet computation mode (mxnet.symbol or mxnet.ndarray)
            :param kws: the dict of input and output variables of the distribution
            :type kws: {name: MXNet NDArray or MXNet Symbol}
            :returns: log pdf of the distribution
            :rtypes: MXNet NDArray or MXNet Symbol
            """
            variables = {name: kws[name] for name, _ in self.inputs}
            variables['random_variable'] = kws['random_variable']
            rv_shape = variables['random_variable'].shape[1:]

            num_samples = max([get_num_samples(F, v) for v in variables.values()])
            full_shape = (num_samples,) + rv_shape

            variables = {name: broadcast_to_w_samples(F, v, full_shape) for
                         name, v in variables.items()}
            res = func(self, F=F, **variables)
            return res
        return log_pdf_broadcast


class UnivariateDrawSamplesDecorator(DrawSamplesDecorator):

    def _wrap_draw_samples_with_broadcast(self, func):
        def draw_samples_broadcast(self, F, rv_shape, num_samples=1,
                                   always_return_tuple=False, **kws):
            """
            Draw a number of samples from the distribution.

            :param F: the MXNet computation mode (mxnet.symbol or mxnet.ndarray)
            :param rv_shape: the shape of each sample
            :type rv_shape: tuple
            :param nSamples: the number of drawn samples (default: one)
            :int nSamples: int
            :param always_return_tuple: Whether return a tuple even if there is only one variables in outputs.
            :type always_return_tuple: boolean
            :param kws: the dict of input variables of the distribution
            :type kws: {name: MXNet NDArray or MXNet Symbol}
            :returns: a set samples of the distribution
            :rtypes: MXNet NDArray or MXNet Symbol or [MXNet NDArray or MXNet Symbol]
            """
            rv_shape = list(rv_shape.values())[0]
            variables = {name: kws[name] for name, _ in self.inputs}

            isSamples = any([is_sampled_array(F, v) for v in variables.values()])
            if isSamples:
                num_samples_inferred = max([get_num_samples(F, v) for v in
                                           variables.values()])
                if num_samples_inferred != num_samples:
                    raise InferenceError("The number of samples in the nSamples argument of draw_samples of Gaussian process has to be the same as the number of samples given to the inputs. nSamples: "+str(num_samples)+" the inferred number of samples from inputs: "+str(num_samples_inferred)+".")
            full_shape = (num_samples,) + rv_shape

            variables = {name: broadcast_to_w_samples(F, v, full_shape) for
                         name, v in variables.items()}
            res = func(self, F=F, rv_shape=rv_shape, num_samples=num_samples,
                       **variables)
            if always_return_tuple:
                res = (res,)
            return res
        return draw_samples_broadcast


class UnivariateDistribution(Distribution):
    """
    The base class of a univariate probability distribution. The univariate
    distribution can be defined over a scalar random variable or an array
    of random variables. In case of an array of random variables, the input
    variables are broadcasted to the shape of the output random variable (array).

    :param inputs: the input variables that parameterize the probability
    distribution.
    :type inputs: {name: Variable}
    :param outputs: the random variables drawn from the distribution.
    :type outputs: {name: Variable}
    :param rand_gen: the random generator (default: MXNetRandomGenerator).
    :type rand_gen: RandomGenerator
    :param dtype: the data type for float point numbers.
    :type dtype: numpy.float32 or numpy.float64
    :param ctx: the mxnet context (default: None/current context).
    :type ctx: None or mxnet.cpu or mxnet.gpu
    """
    def __init__(self, inputs, input_names, output_names, outputs=None,
                 rand_gen=None, dtype=None, ctx=None):
        super(UnivariateDistribution, self).__init__(
            inputs=inputs, outputs=outputs,
            input_names=input_names, output_names=output_names,
            rand_gen=rand_gen, dtype=dtype,
            ctx=ctx)

    def _generate_outputs(self, shape=None):
        """
        Set the output variable of the distribution.

        :param shape: the shape of the random distribution.
        :type shape: tuple
        """
        self.outputs = [('random_variable', Variable(value=self, shape=(1,) if
                        shape is None else shape))]



"""Factor module.

.. autosummary::
    :toctree: _autosummary

"""

from copy import copy
from .model_component import ModelComponent


class Factor(ModelComponent):
    """
    A factor represents a relation among multiple variables in a model such as a distribution, a function or a module. It consists of a list of output
    variables and optionally a list of input variables.

    The ``inputs`` and ``outputs`` argument of ``__init__`` holds the input and output of the factor, which are represented in Python dict. The key of a variable in
    the dict is the name of the variable referred in the context of the factor, e.g., the mean and variance of a normal distribution. The value of a
    variable is the reference to the variable in memory. Both input and output variables are accessible as class attributes.

    The ``inputs`` and ``outputs`` argument of ``__init__`` can be:

    * A list of variables
    * An empty list (no input/output variables)
    * None (the input/output variables are not provided yet.)

    Note that the ``outputs`` argument should never be an empty list, as a factor always outputs some variables.

    :param inputs: the input variables of the factor.
    :type inputs: List of tuples of name to node e.g. [('random_variable': Variable y)] or None
    :param outputs: the output variables of the factor.
    :type outputs: List of tuples of name to node e.g. [('random_variable': Variable y)] or None
    """

    def __getattr__(self, value):
        if value.startswith("__"):
            """
            When python copies objects, it begins by checking for ``__setstate__()`` which doesn't exist, so it calls ``__getattr__()``. Our implementation then
            calls the ``self.inputs`` getter before the object is fully prepared because ``__init__()`` never gets called during the copy. This causes an infinite
            recursion to ``__getattr__()``. By skipping magic methods with "__" prefix, we allow the object to initialize correctly during copying.
            """
            raise AttributeError(value)
        if self.inputs is not None:
            for name, node in self.inputs:
                if name == value:
                    return node
        if self.outputs is not None:
            for name, node in self.outputs:
                if name == value:
                    return node
        else:
            raise AttributeError("''%s' object has no attribute '%s'" % (type(self), value))

    def __init__(self, inputs, outputs, input_names, output_names):
        super(Factor, self).__init__()
        self._check_name_conflict(inputs, outputs)
        self.predecessors = inputs if inputs is not None else []
        self.successors = outputs if outputs is not None else []
        self._input_names = input_names if input_names is not None else []
        self._output_names = output_names if output_names is not None else []

    def __repr__(self):
        out_str = str(self.__class__.__name__)
        if self.predecessors is not None:
            out_str = out_str + \
                '(' + ', '.join([str(name) + '=' + str(var) for
                                 name, var in self.predecessors]) + ')'
        return out_str

    def replicate_self(self, attribute_map=None):
        """
        This functions is a copy constructor for the object.
        In order to perform copy construction we first call ``__new__()`` on the class which creates a blank object.
        We then initialize that object using the method's standard init procedures, and do any extra copying of attributes.

        Replicates this Factor, using new inputs, outputs, and a new uuid.
        Used during model replication to functionally replicate a factor into a new graph.

        :param inputs: new input variables of the factor.
        :type inputs: List of tuples of name to node e.g. [('random_variable': Variable y)] or None
        :param outputs: new output variables of the factor.
        :type outputs: List of tuples of name to node e.g. [('random_variable': Variable y)] or None
        """

        replicant = self.__class__.__new__(self.__class__)

        Factor.__init__(replicant, None, None, copy(self.input_names), copy(self.output_names))
        replicant._uuid = self.uuid
        # replicant.replicant = True
        return replicant

    def _check_name_conflict(self, inputs, outputs):
        if inputs is not None and outputs is not None:
            intersect = set([v for _, v in inputs]) & set([v for _, v in outputs])
            if intersect:
                raise RuntimeError("The inputs and outputs variables of " +
                                   self.__class__.__name__ + " have name conflict: " + str(intersect) + ".")

    @property
    def inputs(self):
        """
        Input variables of the factor.

        :returns: Input variables of the factor.
        :rtype: List of tuples of name to node e.g. [('random_variable': Variable y)]
        """
        return self.predecessors

    @inputs.setter
    def inputs(self, inputs):
        """
        Set Input variables of the factor.

        :param inputs: Input variables of the factor.
        :type inputs: List of tuples of name to node e.g. [('random_variable': Variable y)]
        """
        self.predecessors = inputs

    @property
    def outputs(self):
        """
        Return variables of the factor.

        :returns: Output variables of the factor.
        :rtype: List of tuples of name to node e.g. [('random_variable': Variable y)]
        """
        return self.successors

    @outputs.setter
    def outputs(self, outputs):
        """
        Set Output variables of the factor.

        :param outputs: Output variables of the factor.
        :type outputs: List of tuples of name to node e.g. [('random_variable': Variable y)]
        """
        self.successors = outputs

    def set_outputs(self, variables):
        """
        TODO We don't actually support multi-output.
        """
        variables = [variables] if not isinstance(variables, (list, tuple)) else variables
        self.successors = [(name, variable) for name, variable in zip(self.output_names, variables)]

    @property
    def input_names(self):
        """
        Return Input names.
        """
        return self._input_names

    @property
    def output_names(self):
        """
        Return Output names.
        """
        return self._output_names

    def fetch_runtime_inputs(self, params):
        """
        The helper function to fetch the input variables from a set of
        variables according to the UUIDs of the input variables. It returns a
        dictionary of variables at runtime, where the keys are the name of the
        input variables and the values are the MXNet array at runtime. The
        returned dict can be directly passed into runtime functions of factors
        such as eval for functions and log_pdf and draw_samples for
        distributions.

        :param params: the set of variables where the input variables are
        fetched from.
        :type params: {str (UUID): MXNet NDArray or MXNet Symbol}
        :return: a dict of the input variables, where the keys are the name
        of the input variables and the values are the MXNet array at runtime.
        :rtype: {str (kernel name): MXNet NDArray or MXNet Symbol}
        """
        return {n: params[v.uuid] for n, v in self.inputs}

    def fetch_runtime_outputs(self, params):
        """
        The helper function to fetch the output variables from a set of
        variables according to the UUIDs of the output variables. It returns a
        dictionary of variables at runtime, where the keys are the name of the
        output variables and the values are the MXNet array at runtime. The
        returned dict can be directly passed into runtime functions of factors
        such as eval for functions and log_pdf and draw_samples for
        distributions.

        :param params: the set of variables where the output variables are
        fetched from.
        :type params: {str (UUID): MXNet NDArray or MXNet Symbol}
        :return: a dict of the output variables, where the keys are the name
        of the output variables and the values are the MXNet array at runtime.
        :rtype: {str (kernel name): MXNet NDArray or MXNet Symbol}
        """
        return {n: params[v.uuid] for n, v in self.outputs}



from mxnet.ndarray.ndarray import NDArray
from mxnet.symbol.symbol import Symbol


def add_sample_dimension(F, array):
    """
    Add an extra dimension with shape one in the front (axis 0) of an array repesenting samples.

    :param F: the execution mode of MXNet.
    :type F: mxnet.ndarray or mxnet.symbol
    :param array: the array that the extra dimension is added to.
    :type array: MXNet NDArray or MXNet Symbol
    :returns: the array with the extra dimension.
    :rtypes: the same type as the input array
    """
    return F.expand_dims(array, axis=0)


def add_sample_dimension_arrays(F, arrays):
    """
    Add the sample dimension to a dict of arrays.

    :param F: the execution mode of MXNet.
    :type F: mxnet.ndarray or mxnet.symbol
    :param arrays: a dictionary of MXNet arrays.
    :type arrays: {UUID: array}
    """
    return {
        uuid: add_sample_dimension(F, v) if isinstance(v, (NDArray, Symbol))
        else v for uuid, v in arrays.items()}


def expectation(F, array):
    """
    Return the expectation across the samples if the variable is a set of samples, otherwise return the variable.

    :param F: the MXNet execution mode.
    :type F: mxnet.ndarray or mxnet.symbol
    """
    return F.mean(array, axis=0)


def is_sampled_array(F, array):
    """
    Check if the array is a set of samples.

    :returns: True if the array is a set of samples.
    :rtypes: boolean
    """
    # TODO: replace array.shape with F.shape_array
    return array.shape[0] > 1


def get_num_samples(F, array):
    """
    Get the number of samples in the provided array. If the array is not a set of samples, the return value will be one.

    :returns: the number of samples.
    :rtypes: int
    """
    # TODO: replace array.shape with F.shape_array
    return array.shape[0]


def as_samples(F, array, num_samples):
    """
    Broadcast the variable as if it is a sampled variable. If the variable is already a sampled variable, it directly returns the data reference.

    :param F: the execution mode of MXNet.
    :type F: mxnet.ndarray or mxnet.symbol
    :param array: the array to operate on.
    :type array: MXNet NDArray or MXNet Symbol
    :param num_samples: the number of samples
    :type num_samples: int
    """
    if is_sampled_array(F, array):
        return array
    else:
        return F.broadcast_axis(array, axis=0, size=num_samples)



from abc import abstractmethod
from ..factor import Factor
from ..variables import is_sampled_array, get_num_samples, as_samples
from ..variables import VariableType


class FunctionEvaluationDecorator(object):
    """
    The decorator for the eval function in FunctionEvaluation
    """
    def __call__(self, func):

        func_reshaped = self._wrap_eval_with_reshape(func)
        func_variables = self._wrap_eval_with_variables(func_reshaped)
        return func_variables

    def _wrap_eval_with_variables(self, func):
        def eval_RT_variableset(self, F, variables, always_return_tuple=False):
            """
            The method handling the execution of the function. The inputs arguments of the function are fetched from the *variables* argument according to their UUIDs.

            :param F: the MXNet computation mode (mxnet.symbol or mxnet.ndarray)
            :param variables: the set of MXNet arrays that holds the values of variables at runtime.
            :type variables: {str(UUID): MXNet NDArray or MXNet Symbol}
            :param always_return_tuple: Whether return a tuple even if there is
            only one variables in outputs.
            :type always_return_tuple: boolean
            :returns: the return value of the function
            :rtypes: MXNet NDArray or MXNet Symbol
            """
            args = {name: variables[var.uuid] for name, var in self.inputs
                    if not var.isInherited or var.type == VariableType.RANDVAR}
            return func(
                self, F=F, always_return_tuple=always_return_tuple, **args)
        return eval_RT_variableset

    def _wrap_eval_with_reshape(self, func):
        def eval_RT(self, F, always_return_tuple=False, **input_kws):
            """
            The method handling the execution of the function with RTVariable
            as its input arguments and return values.
            """
            has_samples = any([is_sampled_array(F, v) for v in input_kws.values()])
            if not has_samples:
                # If none of the inputs are samples, directly evaluate the function
                nSamples = 0
                results = func(self, F=F, **{n: v[0] for n, v in input_kws.items()})
                if isinstance(results, (list, tuple)):
                    results = [F.expand_dims(r, axis=0) for r in results]
                else:
                    results = F.expand_dims(results, axis=0)
            else:
                nSamples = max([get_num_samples(F, v) for v in input_kws.values()])
                if self.broadcastable:
                    # If some of the inputs are samples and the function is
                    # broadcastable, evaluate the function with the inputs that are
                    # broadcasted to the right shape.
                    results = func(self, F=F, **{n: as_samples(F, v, nSamples)
                                                 for n, v in input_kws.items()})
                else:
                    # If some of the inputs are samples and the function is *not*
                    # broadcastable, evaluate the function with each set of samples
                    # and concatenate the output variables.
                    results = []
                    for sample_idx in range(nSamples):
                        r = func(
                            self, F=F, **{
                                n: v[sample_idx] if is_sampled_array(F, v) else
                                v[0] for n, v in input_kws.items()})
                        if isinstance(r, (list, tuple)):
                            r = [F.expand_dims(i, axis=0) for i in r]
                        else:
                            r = F.expand_dims(r, axis=0)
                        results.append(r)
                    if isinstance(results[0], (list, tuple)):
                        # if the function has multiple output variables.
                        results = [F.concat([r[i] for r in results], dim=0) for
                                   i in range(len(results[0]))]
                    else:
                        results = F.concat(*results, dim=0)
            results = results if isinstance(results, (list, tuple)) \
                else [results]
            if len(results) == 1 and not always_return_tuple:
                results = results[0]
            return results
        return eval_RT


class FunctionEvaluation(Factor):
    """
    The evaluation of a function with specified input variables.

    :param inputs: the input variables to the function.
    :type inputs:  {variable name : Variable}
    :param outputs: the output variables to the function.
    :type outputs: {variable name : Variable}
    :param broadcastable: Whether the function supports broadcasting with the additional dimension for samples.
    :type: boolean
    """

    def __init__(self, inputs, outputs, input_names, output_names,
                 broadcastable=False):
        self.broadcastable = broadcastable
        super(FunctionEvaluation, self).__init__(
            inputs=inputs, outputs=outputs, input_names=input_names,
            output_names=output_names)

    def replicate_self(self, attribute_map=None):
        replicant = super(FunctionEvaluation, self).replicate_self(attribute_map)
        replicant.broadcastable = self.broadcastable
        return replicant

    @abstractmethod
    @FunctionEvaluationDecorator()
    def eval(self, F, **input_kws):
        """
        The method handling the execution of the function.

        :param F: the MXNet computation mode (mxnet.symbol or mxnet.ndarray)
        :param **input_kws: the dict of inputs to the functions. The key in the
        dict should match with the name of inputs specified in the inputs of
        FunctionEvaluation.
        :type **input_kws: {variable name: MXNet NDArray or MXNet Symbol}
        :returns: the return value of the function
        :rtypes: MXNet NDArray or MXNet Symbol
        """
        raise NotImplementedError
